/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#ifndef _DEBUGMGR_H_
#define _DEBUGMGR_H_

#include "CPU_REG.h"		// kException_Trace

// Types

#pragma mark Types

typedef uae_u32	(*registerFun)(int num);
typedef Bool	(*compareFun)(DWord a, DWord b);

typedef struct BreakpointCondition
// Breakpoint conditions are of the form
//              <register-expr>[<size>] <cond> <value>
// where:
// <register-expr> is either a 68000 register (e.g. "d5") or an indirect reference at a constant
// offset from a 68000 register (e.g. "8(a6)").  An indirect reference reads from memory; when
// you set up a breakpoint condition involving an indirect reference, you must be sure that
// the indirect reference will point to a valid memory address whenever the breakpoint is hit.
//
// <size> is either ".l" (long), ".w" (word) or ".b" (byte).  If no size is specified, the
// expression is assumed to be long.
// <cond> is a binary comparison operator: ==, !=, <=, >=, <, or >
// <value> is a 32-bit integer
// All comparisons are unsigned!
{
	registerFun	regType;
	int			regNum;
	Bool		indirect;
	uae_u32		indirectOffset;
	int			size;	/* number of bytes to compare: 4, 2, or 1 */

	compareFun	condition;
	uae_u32		value;

	// The source text.  We keep this around so that, for example, a user can specify
	// a condition using either hex or decimal notation and they will see the same
	// notation the next time they edit the breakpoint.
	char*		source;

	Bool		Evaluate (void);
} BreakpointCondition;


struct DebugGlobalsType
{
	// Mode settings

	bool			inDebugger;						// true if in debug mode
	bool			ignoreDbgBreaks;				// if true, ignore DbgBreak's
	bool			reEntered;						// true if we re-entered
	bool			firstEntrance;					// true first time we enter debugger
	bool			stepSpy;						// true if step spying.
	bool			breakingOnATrap;				// true if there are A-Traps to check
	bool			continueOverATrap;				// True if skipping over next system call
	bool			continueOverBP;					// True if skipping over next breakpoint

	bool			checkTrapWordOnExit;
	uae_u16			trapWord;
	uae_u16			refNum;

	// Breakpoints and saved opcodes behind each one

	BreakpointType	bp[dbgTotalBreakpoints];
	uae_u16			bpOpcode[dbgTotalBreakpoints];
	BreakpointCondition *bpCondition[dbgTotalBreakpoints];	// condition, or NULL if none

	// Current trap breaks

	Word			trapBreak[dbgTotalTrapBreaks];
	Word			trapParam[dbgTotalTrapBreaks];

	// Step spy support

	uaecptr			ssAddr;							// address to step spy on 
	DWord			ssValue;						// saved value

	// Exception type

	uae_s32			excType;						// why we entered debugger

	// (adam) Data breakpoint support.  This is similar to the step spy capability, but can monitor
	// an arbitrary range of addresses for writes, and doesn't require a saved value.

	bool			watchEnabled;
	uaecptr			watchAddr;		// address to watch, or 0 if none
	DWord			watchBytes;		// number of bytes to watch
};


// For reasons of "convenience", the official type declarations for these
// packets omits their bodies.  For the emulator, we'd really like to have
// the bodies, so we declare these parallel types.

struct SysPktReadRegsRspType2
{
	_sysPktBodyCommon;								// Common Body header
	M68KRegsType	reg;							// <- return registers
};

struct SysPktGetBreakpointsRspType2
{
	_sysPktBodyCommon;								// Common Body header
	BreakpointType	bp[dbgTotalBreakpoints];
};

struct SysPktSetBreakpointsCmdType2
{
	_sysPktBodyCommon;								// Common Body header
	BreakpointType	bp[dbgTotalBreakpoints];
};

struct SysPktGetTrapBreaksRspType2
{
	_sysPktBodyCommon;								// Common Body header
	Word		trapBP[dbgTotalTrapBreaks];
};

struct SysPktSetTrapBreaksCmdType2
{
	_sysPktBodyCommon;								// Common Body header
	Word		trapBP[dbgTotalTrapBreaks];
};

struct SysPktGetTrapConditionsRspType2
{
	_sysPktBodyCommon;								// Common Body header
	Word		trapParam[dbgTotalTrapBreaks];
};

struct SysPktSetTrapConditionsCmdType2
{
	_sysPktBodyCommon;								// Common Body header
	Word		trapParam[dbgTotalTrapBreaks];
};

// class Debug

#pragma mark class Debug

extern DebugGlobalsType		gDebuggerGlobals;
extern Bool 				gFullDebuggerPackets;


class Debug
{
	public:
		static void				Initialize			(void);
		static void				Reset				(void);
		static void				Save				(SessionFile&);
		static void				Load				(SessionFile&);
		static void				Dispose				(void);

		static ErrCode			HandleNewPacket			(const SlkPktHeaderType& header,
														 const SysPktBodyType& pkt);
		static ErrCode			SendPacket				(const void* pkt, long size);
		static ErrCode			SendResponse			(int rsp);
		static ErrCode			SendMessage				(const char*);
		static Bool				BreakIfConnected		(uae_s32 exceptionNumber, uaecptr oldpc);
		static Bool				HandleTrap8				(uae_s32 exceptionNumber, uaecptr oldpc);
		static Bool				HandleSystemCall		(const SystemCallContext& context);
		static ErrCode			EnterDebugger			(uae_s32 reason, uaecptr oldpc);
		static ErrCode			ExitDebugger			(void);
		static Bool				InDebugger				(void);
		static void				CheckStepSpy			(uaecptr writeAddress, int writeBytes);

		static void				HandleCPUBreak			(void);
		static void				InstallCPUBreaks		(void);
		static void				RemoveCPUBreaks			(void);

		static BreakpointCondition*
								NewBreakpointCondition	(const char* sourceString);
		static void				SetBreakpoint			(int index, uaecptr addr, BreakpointCondition* c);
		static void				ClearBreakpoint			(int index);

	private:
		static ErrCode	SendState			(void);
		static ErrCode	ReadMem				(SysPktReadMemCmdType&);
		static ErrCode	WriteMem			(SysPktWriteMemCmdType&);
		static ErrCode	SendRoutineName		(SysPktRtnNameCmdType&);
		static ErrCode	ReadRegs			(SysPktReadRegsCmdType&);
		static ErrCode	WriteRegs			(SysPktWriteRegsCmdType&);
		static ErrCode	Continue			(SysPktContinueCmdType&);
		static ErrCode	RPC					(SysPktRPCType&);
		static ErrCode	GetBreakpoints		(SysPktGetBreakpointsCmdType&);
		static ErrCode	SetBreakpoints		(SysPktSetBreakpointsCmdType2&);
		static ErrCode	ToggleBreak			(SysPktDbgBreakToggleCmdType&);
		static ErrCode	GetTrapBreaks		(SysPktGetTrapBreaksCmdType&);
		static ErrCode	SetTrapBreaks		(SysPktSetTrapBreaksCmdType2&);
		static ErrCode	Find				(SysPktFindCmdType&);
		static ErrCode	GetTrapConditions	(SysPktGetTrapConditionsCmdType&);
		static ErrCode	SetTrapConditions	(SysPktSetTrapConditionsCmdType2&);
		static void		GetRegs				(M68KRegsType&);
		static void		SetRegs				(const M68KRegsType&);
		static void		ConditionalBreak	(uaecptr oldpc);
		static Bool		MustBreakOnTrapSystemCall
											(uae_u16 trapWord, uae_u16 refNum);
};


// This function is called from the RAM setter functions. Make it inline so
// that the common case where stepSpy == FALSE executes quickly.  Yes, this
// _does_ make a difference.

#if 0
#include "Miscellaneous.h"
#include "Bank_SRAM.h"
#include "Gremlins.h"
extern "C" jmp_buf	env;
#endif

inline void Debug::CheckStepSpy (uaecptr writeAddress, int writeBytes)
{
#if 0	// Test code to track down gzip bug.
	if (Gremlins_IsOn ())
	{
		static counter;
		counter++;

		uae_u32		unpackedSize = gRAMBank_Size;
		const void*	image = gRAM_Memory;
		
		// Get the worst-case size fo the compressed data.

		long		worstPackedSize = ::GzipWorstSize (unpackedSize);

		// Create a new buffer to hold the compressed data.

		StMemory	packedImage (worstPackedSize);

		// Get pointers to the source (unpacked) data and
		// destination (packed) data.

		void*	src = (void*) image;
		void*	dest = packedImage.Get ();

		// Compress the data.

		::GzipEncode (&src, &dest, unpackedSize, worstPackedSize);

		// Calculate the compressed size of the data.

		long	packedSize = (char*) dest - (char*) packedImage;

		// Use the chunk size to create a buffer.

		StMemory	unpackedImage (unpackedSize);

		// Get pointers to the source (packed) data and
		// destination (unpacked) data.

		src = packedImage.Get ();
		dest = unpackedImage.Get ();

		// Decompress the data into the dest buffer.

//		if (setjmp (env) == 0)
		if (1)
		{
			::GzipDecode (&src, &dest, packedSize, unpackedSize);
		}
		else
		{
			assert (false);

#if 0
			{
				FILE*	f = fopen ("c:\\temp\\fred.big", "wb");
				if (f)
				{
					fwrite (gRAM_Memory, gRAMBank_Size, 1, f);
					fclose (f);
				}
			}

			{
				FILE*	f = fopen ("c:\\temp\\fred.small", "wb");
				if (f)
				{
					fwrite (packedImage.Get (), packedSize, 1, f);
					fclose (f);
				}
			}
#endif

#if 0
			uae_u32		unpackedSize = gRAMBank_Size;
			const void*	image = gRAM_Memory;
			
			// Get the worst-case size fo the compressed data.

			long		worstPackedSize = ::GzipWorstSize (unpackedSize);

			// Create a new buffer to hold the compressed data.

			StMemory	packedImage (worstPackedSize);

			// Get pointers to the source (unpacked) data and
			// destination (packed) data.

			void*	src = (void*) image;
			void*	dest = packedImage.Get ();

			// Compress the data.

			::GzipEncode (&src, &dest, unpackedSize, worstPackedSize);

			// Calculate the compressed size of the data.

			long	packedSize = (char*) dest - (char*) packedImage;

			// Use the chunk size to create a buffer.

			StMemory	unpackedImage (unpackedSize);

			// Get pointers to the source (packed) data and
			// destination (unpacked) data.

			src = packedImage.Get ();
			dest = unpackedImage.Get ();

			// Decompress the data into the dest buffer.

			if (setjmp (env) == 0)
			{
				::GzipDecode (&src, &dest, packedSize, unpackedSize);
			}
			else
			{
				assert (false);
			}
#endif
		}

		assert (memcmp (image, unpackedImage.Get (), unpackedSize) == 0);
	}
#endif

	if (gDebuggerGlobals.stepSpy)
	{
		if (get_long (gDebuggerGlobals.ssAddr) != gDebuggerGlobals.ssValue)
		{
			// !!! Look into adam's comments below.
			Emulator::SetBreakReason (kException_Trace);
		}
	}
	if (gDebuggerGlobals.watchEnabled &&
		writeAddress < gDebuggerGlobals.watchAddr + gDebuggerGlobals.watchBytes &&
		writeAddress + writeBytes > gDebuggerGlobals.watchAddr)
	{
		// (adam) The step spy code above apparently thinks it can break into
		// the debugger by calling Emulator::SetBreakReason, but that won't work
		// (I tried it).  Software::ProcessException() will do everything we need
		// here; it calls Debug::BreakIfConnected(), which in turn calls
		// Debug::EnterDebugger().
		// We must first check whether we are already in the debugger,
		// because it's possible that a single instruction will trigger the
		// data breakpoint several times.  This can happen, for example, when we hit
		// a trap #15 calling the PalmOS system function MemMove.  In this case, for
		// efficiency, the debugger emulates the entire MemMove call - this happens
		// in the function Headpatch::MemMove.  A single MemMove can trigger the
		// data breakpoint many times, but we must call Debug::EnterDebugger() only once
		// or very bad things will happen.

		/* (adam) old code - this doesn't work because the PC hasn't been incremented yet
		 * if (!fgDebuggerGlobals.inDebugger)
		 *  Software::ProcessException (kException_Trace, m68k_getpc (), m68k_getpc ()); */

		regs.spcflags |= SPCFLAG_DOTRACE;	// break immediately
	}
}


#endif /* _DEBUGMGR_H_ */
  
  
