/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "SocketMessaging.h"

#include "Logging.h"


// ---------------------------------------------------------------------------
//		 Stuff
// ---------------------------------------------------------------------------

enum { kSocketState_Unconnected, kSocketState_Listening, kSocketState_Connected };

static int				gSocketState		= kSocketState_Unconnected;
static SOCKET			gListeningSocket	= INVALID_SOCKET;
static SOCKET			gDebuggerSocket		= INVALID_SOCKET;
static Socket::ConnectCallback
						gConnectCallback	= NULL;

static SOCKET			PrvNewSocket (void);
static sockaddr*		PrvFillAddress (sockaddr* addr, bool forConnect);
static void				PrvErrorOccurred (void);
static int				PrvGetError (void);

#define PRINTF	if (!LogLLDebugger ()) ; else LogAppendMsg


/***********************************************************************
 *
 * FUNCTION:	Socket::Start
 *
 * DESCRIPTION:	Starts up the sockets mechanism for communicating with
 *				the debugger.  It first attempts to connect to the
 *				debugger, assuming that it's been started up and issued
 *				a listen().  If that fails, it creates a socket to
 *				listen on itself.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 *				On any error, the sub-system will be closed down.  All
 *				sockets will be closed and the state will be set to
 *				unconnected.
 *
 ***********************************************************************/

void Socket::Start (ConnectCallback fn)
{
	assert (gSocketState == kSocketState_Unconnected);
	assert (gListeningSocket == INVALID_SOCKET);
	assert (gDebuggerSocket == INVALID_SOCKET);

	gConnectCallback = fn;

	int			result;
	sockaddr	addr;

	gDebuggerSocket = PrvNewSocket ();

	// Attempt to connect to that address (in case anyone is listening).

	result = connect (gDebuggerSocket, PrvFillAddress (&addr, true), sizeof(addr));
	if (result == 0)
	{
		PRINTF ("Socket::Start: connected!");

		gSocketState = kSocketState_Connected;
		return;
	}

	// Didn't connect; attempt to bind to that address so that we
	// can listen for connections.

	closesocket (gDebuggerSocket);
	gDebuggerSocket = INVALID_SOCKET;

	gListeningSocket = PrvNewSocket ();

	result = bind (gListeningSocket, PrvFillAddress (&addr, false), sizeof(addr));
	if (result != 0)
	{
		PRINTF ("bind failed; result = %08X", PrvGetError ());

		PrvErrorOccurred ();
		return;
	}

	// Start listening for a connection.

	result = listen (gListeningSocket, 1);
	if (result != 0)
	{
		PRINTF ("listen failed; result = %08X", PrvGetError ());

		PrvErrorOccurred ();
		return;
	}

	PRINTF ("Socket::Start: listening for connection...");

	gSocketState = kSocketState_Listening;
}


/***********************************************************************
 *
 * FUNCTION:	Socket::Stop
 *
 * DESCRIPTION:	Stops the sockets session.  Closes all sockets, sets
 *				our state to unconnected, and closes down the sockets
 *				library (on Windows).
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void Socket::Stop (void)
{
	if (gDebuggerSocket != INVALID_SOCKET)
	{
		closesocket (gDebuggerSocket);
		gDebuggerSocket = INVALID_SOCKET;
	}

	if (gListeningSocket != INVALID_SOCKET)
	{
		closesocket (gListeningSocket);
		gListeningSocket = INVALID_SOCKET;
	}

	gSocketState = kSocketState_Unconnected;
}


/***********************************************************************
 *
 * FUNCTION:	Socket::ConnectPending
 *
 * DESCRIPTION:	Returns whether or not there is a pending connect.  This
 *				function can be called at any time.  It will only return
 *				true if there is a non-blocking listen issued with a
 *				pending connect that we haven't yet accepted.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	True if there is a pending connect.  If there is, the
 *				caller should next call Socket::AcceptConnection.
 *
 ***********************************************************************/

Bool Socket::ConnectPending (void)
{
	// From the sockets manual for the select() function:
	//
	// The parameter readfds identifies the sockets that are to be
	// checked for readability. If the socket is currently in the
	// listen state, it will be marked as readable if an incoming
	// connection request has been received such that an accept is
	// guaranteed to complete without blocking. For other sockets,
	// readability means that queued data is available for reading
	// such that a call to recv, WSARecv, WSARecvFrom, or recvfrom
	// is guaranteed not to block.

	Bool	havePendingConnect = false;

	if (gSocketState == kSocketState_Unconnected)
	{
		Start (gConnectCallback);
	}

	if (gSocketState == kSocketState_Listening)
	{
		assert (gListeningSocket != INVALID_SOCKET);
		assert (gDebuggerSocket == INVALID_SOCKET);

		fd_set	readfds;

		FD_ZERO (&readfds);
		FD_SET (gListeningSocket, &readfds);

		timeval	timeout = {0, 0};	// Don't wait at all

		int	result = select (0, &readfds, NULL, NULL, &timeout);

		if (result > 0 && FD_ISSET (gListeningSocket, &readfds))
		{
			PRINTF ("Socket::ConnectPending: have pending connection");

			havePendingConnect = true;
		}
		else if (result == SOCKET_ERROR)
		{
			PRINTF ("Socket::ConnectPending: error calling select: %08X", PrvGetError ());

			result = PrvGetError ();
		}
	}

	return havePendingConnect;
}


/***********************************************************************
 *
 * FUNCTION:	Socket::AcceptConnection
 *
 * DESCRIPTION:	Accepts a pending connection.  Assumes that
 *				ConnectPending has already been called and returned
 *				true.  After this, the socket connection should be open
 *				for business.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 *				On any error, the sub-system will be closed down.  All
 *				sockets will be closed and the state will be set to
 *				unconnected.
 *
 ***********************************************************************/

void Socket::AcceptConnection (void)
{
	assert (gSocketState == kSocketState_Listening);
	assert (gListeningSocket != INVALID_SOCKET);
	assert (gDebuggerSocket == INVALID_SOCKET);

	sockaddr	addr;
	int			addr_len;

	memset (&addr, 0, sizeof (addr));
	addr_len = sizeof(addr);

	gDebuggerSocket = accept (gListeningSocket, &addr, &addr_len);

	if (gDebuggerSocket == INVALID_SOCKET)
	{
		PRINTF ("Socket::AcceptConnection: error calling accept: %08X", PrvGetError ());

		PrvErrorOccurred ();
		return;
	}

	closesocket (gListeningSocket);
	gListeningSocket = INVALID_SOCKET;

	// Set the socket to not use the "Nagle delay algorithm" that clumps
	// batches of small writes together into one big send.

	protoent*	protocolEntity = getprotobyname ("tcp");
	if (!protocolEntity)
	{
		PRINTF ("Socket::AcceptConnection: error calling getprotobyname: %08X", PrvGetError ());

		PrvErrorOccurred ();
		return;
	}

	int	tmp = 1;
	if (setsockopt (gDebuggerSocket, protocolEntity->p_proto, TCP_NODELAY, (char*) &tmp, sizeof (tmp)))
	{
		PRINTF ("Socket::AcceptConnection: error calling setsockopt: %08X", PrvGetError ());

		PrvErrorOccurred ();
		return;
	}

	PRINTF ("Socket::AcceptConnection: accepted the connection.");

	gSocketState = kSocketState_Connected;

	if (gConnectCallback)
		gConnectCallback ();
}


/***********************************************************************
 *
 * FUNCTION:	Socket::IsConnected
 *
 * DESCRIPTION:	Returns whether or not we have a live connection to an
 *				external debugger via a socket.  This function can be
 *				called at any time.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 *				On any error, the sub-system will be closed down.  All
 *				sockets will be closed and the state will be set to
 *				unconnected.
 *
 ***********************************************************************/

Bool Socket::IsConnected (void)
{
	// If we thought we were connected, see if we still are.

	if (gSocketState == kSocketState_Connected)
	{
		// From the sockets manual for the select() function:
		//
		// If a socket is processing a connect call (nonblocking), a socket is
		// writable if the connection establishment successfully completes. If
		// the socket is not processing a connect call, writability means a send,
		// sendto, or WSASendto are guaranteed to succeed.

		fd_set	writefds;

		FD_ZERO (&writefds);
		FD_SET (gDebuggerSocket, &writefds);

		timeval	timeout = {0, 0};	// Don't wait at all

		int		result = select (0, NULL, &writefds, NULL, &timeout);

		Bool	isConnected = result > 0 && FD_ISSET (gDebuggerSocket, &writefds);

		// If we're no longer connected, update our internal state.

		if (!isConnected)
		{
			PRINTF ("Socket::IsConnected: found a broken connection, or error calling select: %08X", PrvGetError ());

			PrvErrorOccurred ();
		}
	}

	return gSocketState == kSocketState_Connected;
}


/***********************************************************************
 *
 * FUNCTION:	Socket::CheckConnection
 *
 * DESCRIPTION:	Attempt to ensure that a connection is in place.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	True if a connection is in place.  False otherwise.
 *
 ***********************************************************************/

Bool Socket::CheckConnection (void)
{
	if (!IsConnected() && ConnectPending ())
	{
		AcceptConnection ();
	}

	return IsConnected ();
}


/***********************************************************************
 *
 * FUNCTION:	Socket::HasUnreadData
 *
 * DESCRIPTION:	Returns whether or not there is data to be read from
 *				the socket.
 *
 * PARAMETERS:	msecs - time to wait in milliseconds for any data.
 *
 * RETURNED:	True if there is data to be read.  There is no way to
 *				get how many bytes are currently in the buffer.
 *
 *				On any error, the sub-system will be closed down.  All
 *				sockets will be closed and the state will be set to
 *				unconnected.
 *
 ***********************************************************************/

Bool Socket::HasUnreadData (long msecs)
{
	Bool	hasData = false;

	// From the sockets manual for the select() function:
	//
	// The parameter readfds identifies the sockets that are to be
	// checked for readability. If the socket is currently in the
	// listen state, it will be marked as readable if an incoming
	// connection request has been received such that an accept is
	// guaranteed to complete without blocking. For other sockets,
	// readability means that queued data is available for reading
	// such that a call to recv, WSARecv, WSARecvFrom, or recvfrom
	// is guaranteed not to block.

	if (CheckConnection ())
	{
		assert (gSocketState == kSocketState_Connected);

		fd_set	readfds;

		FD_ZERO (&readfds);
		FD_SET (gDebuggerSocket, &readfds);

		timeval	timeout = {msecs / 1000, msecs * 1000};

		int	result = select (0, &readfds, NULL, NULL, &timeout);

		if (result > 0 && FD_ISSET (gDebuggerSocket, &readfds))
		{
			PRINTF ("Socket::HasUnreadData: has pending data.");
			hasData = true;
		}
		else if (result == SOCKET_ERROR)
		{
			PrvErrorOccurred ();
			return false;
		}
	}

	return hasData;
}


/***********************************************************************
 *
 * FUNCTION:	Socket::Write
 *
 * DESCRIPTION:	Writes data out to the buffer.
 *
 * PARAMETERS:	buffer - pointer to the data to be written.
 *
 *				amountToWrite - number of bytes in the buffer.
 *
 * RETURNED:	Nothing.
 *
 *				On any error, the sub-system will be closed down.  All
 *				sockets will be closed and the state will be set to
 *				unconnected.
 *
 ***********************************************************************/

long Socket::Write (const void* buffer, long amountToWrite)
{
	int amountWritten = 0;

	if (CheckConnection ())
	{
		if (LogLLDebuggerData ())
			LogAppendData (buffer, amountToWrite, "Socket::Write: writing %ld bytes.", amountToWrite);
		else
			PRINTF ("Socket::Write: writing %ld bytes.", amountToWrite);

		while (amountToWrite)
		{
			int	sentThisTime = send (gDebuggerSocket,
									&((char*) buffer)[amountWritten],
									amountToWrite, 0);

			if (sentThisTime == SOCKET_ERROR )
			{
				PrvErrorOccurred ();
				return amountWritten;
			}

			amountWritten += sentThisTime;
			amountToWrite -= sentThisTime;
		}
	}
	else
	{
		PRINTF ("Socket::Write: no connection.");
	}

	return amountWritten;
}


/***********************************************************************
 *
 * FUNCTION:	Socket::Read
 *
 * DESCRIPTION:	Read bytes from the socket.  At most sizeOfBuffer bytes
 *				are read.  It is assumed that HasUnreadData has already
 *				been called and returned true.
 *
 * PARAMETERS:	buffer - pointer to the buffer to put the data.
 *
 *				sizeOfBuffer - size of the buffer.
 *
 * RETURNED:	The actual number of bytes read..
 *
 *				On any error, the sub-system will be closed down.  All
 *				sockets will be closed and the state will be set to
 *				unconnected.
 *
 ***********************************************************************/

long Socket::Read (void* buffer, long sizeOfBuffer)
{
	long	amountRead = 0;

	if (CheckConnection ())
	{
		while (amountRead < sizeOfBuffer)
		{
			long	r = recv (gDebuggerSocket, ((char*) buffer) + amountRead, sizeOfBuffer - amountRead, 0);

			if (r == SOCKET_ERROR)
			{
				PrvErrorOccurred ();
				return amountRead;
			}

			if (r > 0)
			{
				amountRead += r;

				if (LogLLDebuggerData ())
					LogAppendData (((char*) buffer) + amountRead - r, r,
								"Socket::Read: Got %ld of %ld bytes of data.", amountRead, sizeOfBuffer);
				else
					PRINTF ("Socket::Read: Got %ld of %ld bytes of data.", amountRead, sizeOfBuffer);
			}
		}
	}
	else
	{
		PRINTF ("Socket::Read: no connection.");
	}

	return amountRead;
}


/***********************************************************************
 *
 * FUNCTION:	PrvNewSocket
 *
 * DESCRIPTION:	Creates and returns a new socket configured the way we
 *				like it.  The way we like it is: reusable after it's
 *				closed, uses "keep alive" packets, and non-blocking.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	The created socket, or INVALID_SOCKET on error.
 *
 *				On any error, the sub-system will be closed down.  All
 *				sockets will be closed and the state will be set to
 *				unconnected.
 *
 ***********************************************************************/

static SOCKET PrvNewSocket (void)
{
	int		tmp;
	int		result;

	SOCKET	newSocket = socket (AF_INET, SOCK_STREAM, 0);
	if (newSocket == INVALID_SOCKET)
	{
		goto Error;
	}

	// Allow rapid reuse of this port.

	tmp = 1;
	result = setsockopt (newSocket, SOL_SOCKET, SO_REUSEADDR, (char*) &tmp, sizeof(tmp));
	if (result != 0)
	{
		goto Error;
	}

	// Enable TCP keep alive process.

	tmp = 1;
	result = setsockopt (newSocket, SOL_SOCKET, SO_KEEPALIVE, (char*) &tmp, sizeof(tmp));
	if (result != 0)
	{
		goto Error;
	}

	return newSocket;

Error:
	if (newSocket != INVALID_SOCKET)
	{
		closesocket (newSocket);
	}

	int	error = PrvGetError ();
	return INVALID_SOCKET;
}


/***********************************************************************
 *
 * FUNCTION:	PrvFillAddress
 *
 * DESCRIPTION:	Fill in a sockaddr data structure with values
 *				appropriate for connecting.  For now, this is an
 *				Internet address on the local machine with port 2000.
 *
 * PARAMETERS:	addr - the sockaddr to fill in.
 *
 * RETURNED:	The same sockaddr passed in.
 *
 ***********************************************************************/

sockaddr* PrvFillAddress (sockaddr* addr, bool forConnect)
{
	const long		kDefaultPortNumber = 2000;
	sockaddr_in*	addr_in = (sockaddr_in*) addr;

	addr_in->sin_family			= AF_INET;
	addr_in->sin_port			= htons (kDefaultPortNumber);
	addr_in->sin_addr.s_addr	= htonl (forConnect ? INADDR_LOOPBACK : INADDR_ANY);

	return addr;
}


/***********************************************************************
 *
 * FUNCTION:	PrvErrorOccurred
 *
 * DESCRIPTION:	Handles any internal errors.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void PrvErrorOccurred (void)
{
	int	result = PrvGetError ();
	Socket::Stop ();
}


/***********************************************************************
 *
 * FUNCTION:	PrvGetError
 *
 * DESCRIPTION:	Returns the error code for the last operation.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	The error code as returned by WSAGetLastError (on
 *				Windows) or errno (on the Mac or UNIX).
 *
 ***********************************************************************/

int PrvGetError (void)
{
#if defined (_WIN32)
	return ::WSAGetLastError ();
#else
	return errno;
#endif
}
