/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#ifndef __DOCUMENT_H__
#define __DOCUMENT_H__

#include "CPU_MT.h"
#include "AppPreferences.h"

// -----------------------------------------------------------------------------
// class Document
//
// The class encapsulates all of the DATA pertinent to an emulation session.
// Some of this data is persistent, and this class provides methods for saving
// that data.
// In the future, it would be nice if each of the data member of this class
// (in particular the CPU) would be thread safe, so that I could have several
// Documents open at once.
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// local types
// -----------------------------------------------------------------------------

typedef void (*PropertyChangedHandler)( int property, void *param );

typedef struct
{
    PropertyChangedHandler func;
    void *param;
} PropertyChangedHandlerSpec;

class Document
{

  public:

    // -----------------------------------------------------------------------------
    // public constants
    // -----------------------------------------------------------------------------

    enum {
        unspecifiedChanged = 0,
        documentOpening,
        documentClosing,
        cpuStatusChanged
    };
    
    // -----------------------------------------------------------------------------
    // constructor / destructor
    // -----------------------------------------------------------------------------
    
    Document( const AppPreferences::Configuration& cfg );
    virtual ~Document( void );

    // -----------------------------------------------------------------------------
    // public methods
    // -----------------------------------------------------------------------------

    ErrCode open( FileReference& f );

    void emulatorStart( void );
    void emulatorStop( void );
    void emulatorReset( void );
    void loadApplication( FileReference& appFile );
    void saveMemory();
    void saveMemoryAs( FileReference& dest );
    void saveScreen( FileReference& dest );

    void hotsync();
    void pen( bool down, int x, int y ) { _cpu->pen( down, x, y ); }
    void button( bool down, int button ) { _cpu->button( down, button ); }
    void putkey( unsigned char key ) { _cpu->putkey( key ); }

    FileReference getCurrentRAMRef() { return _currentRAMRef; }

    void addPropertyChangedHandler( PropertyChangedHandler callback, void *param );
    void removePropertyChangedHandler( PropertyChangedHandler callback );

    bool LcdIsOn();

    bool initializedOk() { return _cpu != NULL; }
    AppPreferences::Configuration& getCurrentConfiguration() { return _currentConfiguration; }

    void gremlinsNew();
    void gremlinsStep();
    void gremlinsResume();
    void gremlinsStop();

  private:
    
    // -----------------------------------------------------------------------------
    // private methods
    // -----------------------------------------------------------------------------

    void propertyChanged( int property = unspecifiedChanged );
    
    FileReference findPairedROMFile( FileHandle& docFile );
    void saveRAMImage( const FileReference&, Bool forAutoSave );

    // -----------------------------------------------------------------------------
    // private friends
    // -----------------------------------------------------------------------------

    friend void _hotSyncUp( void *data );
	friend void Platform::AutoSaveState (const FileReference&);	// saveRAMImage

    // -----------------------------------------------------------------------------
    // private data members
    // -----------------------------------------------------------------------------
    
    CPU *_cpu;
    AppPreferences::Configuration _currentConfiguration;
    FileReference _currentRAMRef;
    list<PropertyChangedHandlerSpec*> _handlers;

};


#endif /* __DOCUMENT_H__ */
