/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"

#include "SessionFile.h"

#include <unistd.h>
#include <sys/time.h>
#include <sys/stat.h>			// mkdir
#include <time.h>
#include <ctype.h>

#include "omnithread.h"
#include "ResStrings.h"
#include "Gremlins.h"			// Gremlins_EventCounter
#include "Platform.h"
#include "fltk_Dialog.h"
#include "fltk_MainWindow.h"	// MainWindow::getMainWindow


// ===========================================================================
//		 Globals
// ===========================================================================

unsigned long gPseudoTickcount;
static time_t gStartTime;
static Bool gForceNewDirectory;

// ===========================================================================
//		 TByteQueuePrivate
// ===========================================================================

// Queue used for keyboard and serial input/output.  This is the
// private, platform-dependent implementation used by TByteQueue.

class TByteQueuePrivate
{
public:
    TByteQueuePrivate();
    ~TByteQueuePrivate();
  
    void Put( uae_u8 b );
    uae_u8 Get( void );
    int GetUsed( void ) { return _usedCount; }
    int GetFree( void ) { return _freeCount; }

    Bool WaitForDataAvailable( int timeoutms );
  
private:
  
    enum { QUEUE_SIZE = 8 };
  
    long _usedCount;
    long _freeCount;
    uae_u8 _buffer[QUEUE_SIZE];
    int _head;
    int _tail;

    omni_mutex _mutex;
    omni_condition _available;
};

// ===========================================================================
//		 Platform
// ===========================================================================

// Compare lexigraphically two strings

int _stricmp( const char *s1, const char *s2 )
{
    return strcasecmp( s1, s2 );
}

int _strnicmp( const char *s1, const char *s2, int n )
{
    return strncasecmp( s1, s2, n );
}

char* _strdup( const char *s )
{
    return strdup( s );
}

// ---------------------------------------------------------------------------
//		 Platform::Initialize
// ---------------------------------------------------------------------------
// Initializes platform-dependent stuff.

void Platform::Initialize( void )
{
    gStartTime = time( NULL );
}


// ---------------------------------------------------------------------------
//		 Platform::Reset
// ---------------------------------------------------------------------------

void Platform::Reset( void )
{
    gStartTime = time( NULL );
}


// ---------------------------------------------------------------------------
//		 Platform::Save
// ---------------------------------------------------------------------------

void Platform::Save(SessionFile&)
{
}


// ---------------------------------------------------------------------------
//		 Platform::Load
// ---------------------------------------------------------------------------

void Platform::Load(SessionFile&)
{
}


// ---------------------------------------------------------------------------
//		 Platform::Dispose
// ---------------------------------------------------------------------------

void Platform::Dispose( void )
{
}


// ---------------------------------------------------------------------------
//		 Platform::CommonDialog
// ---------------------------------------------------------------------------

int Platform::CommonDialog( const char* title, const char* msg, int flags )
{
    return fltkDialog( title, msg, flags );
}


// ---------------------------------------------------------------------------
//		 Platform::GetString
// ---------------------------------------------------------------------------

string Platform::GetString( int id )
{
    return _ResGetString( id );
}


// ---------------------------------------------------------------------------
//		 Platform::GetIDForError
// ---------------------------------------------------------------------------

int Platform::GetIDForError( ErrCode error )
{
#if 0
	return kStr_GenericError;
#endif
	return 0;
}


// ---------------------------------------------------------------------------
//		 Platform::GetIDForRecovery
// ---------------------------------------------------------------------------

int Platform::GetIDForRecovery( ErrCode error )
{
	return 0;
}


// ---------------------------------------------------------------------------
//		 Platform::GetShortVersionString
// ---------------------------------------------------------------------------
// Returns a short version string.  The format of the string is:
//
//		#.# (.#) ([dab]#)
//
//		# = one or more numeric digits
//		. = literal '.'
//		Patterns in parentheses are optional
//		Patterns in brackets mean "one of these characters"
//		Spaces are shown above for clarity; they do not appear in the string
//
// Valid strings would be: 2.1d7, 2.1.1b14, 2.99, 2.1.1

string Platform::GetShortVersionString( void )
{
	return string("Platform::GetShortVersionString not implemented");
}

/***********************************************************************
 *
 * FUNCTION:	ToHostEOL
 *
 * DESCRIPTION:	Converts a string of characters into another string
 *				where the EOL sequence is consistant for files on the
 *				current platform.
 *
 * PARAMETERS:	dest - receives the result.  The buffer is sized by
 *					this function, so the caller doesn't have to
 *					allocate any space itself.
 *
 *				destLen - receives the length of the resulting string.
 *
 *				src - pointer to input characters.
 *
 *				srcLen - number of characters pointed to by src.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void Platform::ToHostEOL (	StMemory& dest, long& destLen,
							const char* src, long srcLen)
{
	char*	d = (char*) Platform::AllocateMemory (srcLen);
	char*	p = d;
	Bool	previousWas0x0D = false;

	for (long ii = 0; ii < srcLen; ++ii)
	{
		char	ch = src[ii];

		// Convert 0x0D to 0x0A.
		
		if (ch == 0x0D)
		{
			*p++ = 0x0A;
		}

		// If we're looking at a 0x0A that's part of
		// a 0x0D/0x0A, just swallow it.

		else if (ch == 0x0A && previousWas0x0D)
		{
			// Nothing
		}

		// Copy all other characters straight through.

		else
		{
			*p++ = ch;
		}

		previousWas0x0D = ch == 0x0D;
	}

	destLen = p - d;
	Platform::ReallocMemory (d, destLen);
	dest.Adopt (d);
}


/***********************************************************************
 *
 * FUNCTION:	Platform::CreateReferenceInEmulatorPath
 *
 * DESCRIPTION:	Create a file reference to a file with the given name
 *				and that resides in the same directory as the emulator.
 *
 * PARAMETERS:	fileName - Name of the file the caller wants to create.
 *
 * RETURNED:	The desired FileReference.  This object could be
 *				invalid if an error occured.
 *
 ***********************************************************************/

FileReference Platform::CreateReferenceInEmulatorPath (const char* fileName)
{
	char*		home = getenv ("HOME");
	char*		path = (char*) malloc (strlen (home) + 1 + strlen (fileName) + 1);

	strcpy (path, home);
	strcat (path, "/");
	strcat (path, fileName);

	FileReference result (path);

	free (path);

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	Platform::CreateReferenceInGremlinStatePath
 *
 * DESCRIPTION:	Create a file reference to a file with the given name
 *				and that resides in the same directory designated to
 *				hold auto-saved Gremlin states.
 *
 *				If gForceNewDirectory is true, a directory is created
 *				and used.
 *
 * PARAMETERS:	fileName - Name of the file the caller wants to create.
 *
 * RETURNED:	The desired FileReference.  This object could be
 *				invalid if an error occured.
 *
 ***********************************************************************/

FileReference Platform::CreateReferenceInGremlinStatePath (const char* fileName)
{
	FileReference result;
	
	// If requested, create the directory to use.

	static char	gGremlinStatePath[PATH_MAX];

	if (gForceNewDirectory)
	{
		gForceNewDirectory = false;
		
		sprintf (gGremlinStatePath, "%s/GremlinStates_%ld",
				 getenv ("HOME"),
				 Platform::GetMilliseconds ());

		if (mkdir (gGremlinStatePath, 0777) != 0)
		{
			// !!! Put up some error message.
			
			gGremlinStatePath[0] = 0;
		}
	}

	// If we don't have a directory, blow out.

	if (gGremlinStatePath[0] == 0)
	{
		return result;
	}

	// Create the file reference from the full path.

	char		fullPath[PATH_MAX];
	sprintf (fullPath, "%s/%s", gGremlinStatePath, fileName);

	result = FileReference (fullPath);

	return result;
}


// ---------------------------------------------------------------------------
//		 Platform::GetLoadableFileList
// ---------------------------------------------------------------------------

void Platform::GetLoadableFileList( string directoryName, LoadableList& fileList )
{
#if 0
	// Get the application's directory.

	char	path[_MAX_PATH], drive[_MAX_DRIVE], dir[_MAX_DIR];

	::GetModuleFileName (gInstance, path, _MAX_PATH);
	_splitpath (path, drive, dir, NULL, NULL);
	_makepath (path, drive, dir, NULL, NULL);

	string	pathName (path + directoryName + "\\");
	string	searchPattern (pathName + "*");

	// Start iterating over it.  We'll iterate over all the contents
	// and filter by filetype by hand.

	WIN32_FIND_DATA	fileData;
	HANDLE			hSearch = ::FindFirstFile (searchPattern.c_str (), &fileData);
	if (hSearch == INVALID_HANDLE_VALUE)
	{
		// If an error occurs (probably no such directory), just return.

		return;
	}


	BOOL	finished = FALSE;
	while (!finished)
	{
		if (::EndsWith (fileData.cFileName, ".prc") ||
			::EndsWith (fileData.cFileName, ".pdb") ||
			::EndsWith (fileData.cFileName, ".pqa"))
		{
			string	fullPath (pathName + fileData.cFileName);
			fileList.push_back (LoadFilePB (FileReference (fullPath)));
		}

		if (!::FindNextFile (hSearch, &fileData))
		{
			DWORD	lastError = GetLastError ();
			if (lastError != ERROR_SUCCESS)
			{
				// Blow out of here on *any* error, not just ERROR_NO_MORE_FILES.
				finished = TRUE;
			}
		}
	}

	::FindClose (hSearch);
#endif
}

// -----------------------------------------------------------------------------
// find the ROM file path embedded in the saved ram image
Bool Platform::ReadROMFileReference (ChunkFile& docFile, FileReference& f)
{
	// First look for a ROM file path.

	string	path;
	if (docFile.ReadString (SessionFile::kROMUnixPathTag, path))
	{
		f = FileReference (path);
		return true;
	}

	// If a path can't be found, look for a simple ROM name.

	string	name;
	if (docFile.ReadString (SessionFile::kROMNameTag, name))
	{
		// !!! TBD
	}

	return false;
}

void Platform::WriteROMFileReference (ChunkFile& docFile, const FileReference& f)
{
	docFile.WriteString (SessionFile::kROMUnixPathTag, f.GetFilePath ());
}


// ---------------------------------------------------------------------------
//		 Platform::UseNewAutoSaveDirectory
// ---------------------------------------------------------------------------


void Platform::UseNewAutoSaveDirectory( void )
{
	gForceNewDirectory = true;
}


// ---------------------------------------------------------------------------
//		 Platform::AutoSaveState
// ---------------------------------------------------------------------------

void Platform::AutoSaveState(const FileReference& fileRef)
{
	MainWindow::getMainWindow()->getDocument()->saveRAMImage (fileRef, true);
}


// ---------------------------------------------------------------------------
//		 Platform::Delay
// ---------------------------------------------------------------------------
// Delay for a time period appropriate for a sleeping 68328.

void Platform::Delay( void )
{
	// Delay 10 msecs.  Delaying by this amount pauses us 1/100 of a second,
	// which is the rate at which the device's tickcount counter increments.
	//
	// Wait on an event instead of just calling Sleep(10) so that another
	// thread can wake us up before our time.

	omni_thread::sleep( 0, 10000 ); // 10k nanoseconds = 1/100 sec
}


// ---------------------------------------------------------------------------
//		 Platform::Cycle
// ---------------------------------------------------------------------------

void Platform::Cycle( void )
{
	// Nothing to do in Unix.
}


// ---------------------------------------------------------------------------
//		 Platform::AllocateMemory
// ---------------------------------------------------------------------------

void* Platform::AllocateMemory( size_t size, Bool clear )
{
	void*	result;

	if (clear)
		result = calloc (size, 1);
	else
		result = malloc (size);

	Errors::ThrowIfNULL (result);

	return result;
}


// ---------------------------------------------------------------------------
//		 Platform::ReallocMemory
// ---------------------------------------------------------------------------

void* Platform::ReallocMemory( void* p, size_t size )
{
	void*	result = realloc (p, size);

	Errors::ThrowIfNULL (result);

	return result;
}


// ---------------------------------------------------------------------------
//		 Platform::RealDisposeMemory
// ---------------------------------------------------------------------------

void Platform::RealDisposeMemory( void* p )
{
	if (p)
	{
		free (p);
	}
}


// ---------------------------------------------------------------------------
//		 Platform::StopOnResetKeyDown
// ---------------------------------------------------------------------------

Bool Platform::StopOnResetKeyDown( void )
{
	// Comment this out for now.  It seems that Windows doesn't always return
	// the expected result.  That is, this function often returns TRUE even
	// though the Control is not down.
	//
	// Since this functionality is really only required by Palm OS engineers,
	// and since they're working on Macs, we don't really need this feature
	// in the Windows version now anyway.

//	return ::GetAsyncKeyState (VK_CONTROL) != 0;
	return false;
}


// ---------------------------------------------------------------------------
//		 Platform::GetMilliseconds
// ---------------------------------------------------------------------------

uae_u32 Platform::GetMilliseconds( void )
{
    struct timeval tv;

    // return milliseconds since reset
    gettimeofday( &tv, NULL );
    uae_u32 millis = (tv.tv_sec - gStartTime) * 1000 + tv.tv_usec/1000;

    return millis;
}



// ---------------------------------------------------------------------------
//		 Platform::HeadPatch
// ---------------------------------------------------------------------------

#if 0
int Platform::HeadPatch( uae_u16 iATrap )
{
	switch (iATrap)
	{
		case sysTrapSndDoCmd:
		{
			SndCommandType*	sc = (SndCommandType*) get_real_address(get_long(m68k_areg(regs, 7) + 4));
			if (do_get_mem_word(&sc->cmd) == 1)
			{
				if (gEnableSounds)
				{
					Beep(do_get_mem_word(&sc->param1lo), do_get_mem_word(&sc->param2));
				}

				return true;
			}

			break;
		}
	}

	return false;
}
#endif


// ---------------------------------------------------------------------------
//		 Platform::ConnectToDebugger
// ---------------------------------------------------------------------------

ErrCode Platform::ConnectToDebugger( void )
{
	// Don't call MMFOpenFiles here.  Doing so would create the
	// hidden window used for message passing in a thread different
	// from the one that processes the messages.

//	::MMFOpenFiles (NULL, HandlePacket, (LPARAM) LCD_GetWindow (),
//					kFromEmulatorFileName, kToEmulatorFileName);

	return 0;
}


// ---------------------------------------------------------------------------
//		 Platform::SendPacket
// ---------------------------------------------------------------------------

ErrCode Platform::SendPacket( const void* pkt, long size )
{
#if 0
	return (uae_u32) ::MMFSendData (pkt, size);
#endif
	return 0;
}


// ---------------------------------------------------------------------------
//		 Platform::DisconnectFromDebugger
// ---------------------------------------------------------------------------

void Platform::DisconnectFromDebugger( void )
{
#if 0
	::MMFCloseFiles (gInstance);
#endif
}


// ---------------------------------------------------------------------------
//		 Platform::ExitDebugger
// ---------------------------------------------------------------------------
//	Perform platform-specific operations when debug mode is exited.

void Platform::ExitDebugger( void )
{
}


// ---------------------------------------------------------------------------
//		 Platform::ViewDrawLine
// ---------------------------------------------------------------------------

void Platform::ViewDrawLine( int xStart, int yStart, int xEnd, int yEnd )
{
}


// ---------------------------------------------------------------------------
//		 Platform::ViewDrawPixel
// ---------------------------------------------------------------------------

void Platform::ViewDrawPixel( int xPos, int yPos )
{
}

// ---------------------------------------------------------------------------
//		 Platform::OpenSerialPort
// ---------------------------------------------------------------------------

ErrCode Platform::OpenSerialPort( void )
{
	return 0;
}


// ---------------------------------------------------------------------------
//		 Platform::SetSerialPortSettings
// ---------------------------------------------------------------------------

ErrCode Platform::SetSerialPortSettings( 	UART::Parity	parity,
								int				stopBits,
								int				dataBits,
								uae_u32			baud,
								Bool			hwHandshaking )
{
	return 0;
}


// ---------------------------------------------------------------------------
//		 Platform::SerialPortOpen
// ---------------------------------------------------------------------------

Bool Platform::SerialPortOpen( void )
{
	return false;
}


// ---------------------------------------------------------------------------
//		 Platform::CloseSerialPort
// ---------------------------------------------------------------------------

void Platform::CloseSerialPort( void )
{
}


// ---------------------------------------------------------------------------
//		 Platform::TransmitTxFIFO
// ---------------------------------------------------------------------------

void Platform::TransmitTxFIFO( void )
{
}


// ---------------------------------------------------------------------------
//		 Platform::ReceiveRxFIFO
// ---------------------------------------------------------------------------

void Platform::ReceiveRxFIFO( void )
{
}


#if 0
// ---------------------------------------------------------------------------
//		 Platform::SetBaud
// ---------------------------------------------------------------------------

void Platform::SetBaud( uae_u32 newRate )
{
}


// ---------------------------------------------------------------------------
//		 Platform::SetCTS
// ---------------------------------------------------------------------------

void Platform::SetCTS( Bool useCTS )
{
}


// ---------------------------------------------------------------------------
//		 Platform::WriteSerialByte
// ---------------------------------------------------------------------------

void Platform::WriteSerialByte( uae_u8 outByte )
{
}


// ---------------------------------------------------------------------------
//		 Platform::ReadSerialByte
// ---------------------------------------------------------------------------

uae_u8 Platform::ReadSerialByte( void )
{
}


// ---------------------------------------------------------------------------
//		 Platform::IncomingSerialBytes
// ---------------------------------------------------------------------------

uae_u32 Platform::IncomingSerialBytes( void )
{
}


// ---------------------------------------------------------------------------
//		 Platform::OutgoingSerialBytes
// ---------------------------------------------------------------------------

uae_u32 Platform::OutgoingSerialBytes( void )
{
}
#endif


#if 0
// ---------------------------------------------------------------------------
//		 CommRead
// ---------------------------------------------------------------------------
// This function sits in its own thread, waiting for data to show up in the
// serial port.  If data arrives, this function plucks it out and stores it
// in a thread-safe queue.

DWORD __stdcall CommRead( void* )
{
}


// ---------------------------------------------------------------------------
//		 CommWrite
// ---------------------------------------------------------------------------
// This function sits in its own thread, waiting for data to show up in the
// write queue.  If data arrives, this function plucks it out and sends it
// out to the port.

DWORD __stdcall CommWrite( void* )
{
}
#endif


// ---------------------------------------------------------------------------
//		 Load File Progress Window
// ---------------------------------------------------------------------------

// ---------------------------------------------------------------------------
//		 Platform::LFPW_Open
// ---------------------------------------------------------------------------

void Platform::LFPW_Open( void )
{
}


// ---------------------------------------------------------------------------
//		 Platform::LFPW_SetNumFiles
// ---------------------------------------------------------------------------

void Platform::LFPW_SetNumFiles( int numFiles )
{
}


// ---------------------------------------------------------------------------
//		 Platform::LFPS_StartNewFile
// ---------------------------------------------------------------------------

void Platform::LFPW_StartNewFile( const char* path )
{
}


// ---------------------------------------------------------------------------
//		 Platform::LFPW_SetFileMax
// ---------------------------------------------------------------------------

void Platform::LFPW_SetFileMax( int max )
{
}


// ---------------------------------------------------------------------------
//		 Platform::LFPW_SetFileValue
// ---------------------------------------------------------------------------

void Platform::LFPW_SetFileValue( int value )
{
}


// ---------------------------------------------------------------------------
//		 Platform::LFPW_Close
// ---------------------------------------------------------------------------

void Platform::LFPW_Close( void )
{
}


// ---------------------------------------------------------------------------
//		 Gremlin Control Window
// ---------------------------------------------------------------------------

// ---------------------------------------------------------------------------
//		 Platform::GCW_Open
// ---------------------------------------------------------------------------

void Platform::GCW_Open( void )
{
}


// ---------------------------------------------------------------------------
//		 Platform::GCW_Close
// ---------------------------------------------------------------------------

void Platform::GCW_Close( void )
{
}


// ===========================================================================
//		 TByteQueue
// ===========================================================================

// ---------------------------------------------------------------------------
//		 TByteQueue::TByteQueue
// ---------------------------------------------------------------------------

TByteQueue::TByteQueue():
        fImplementation( new TByteQueuePrivate )
{
}


// ---------------------------------------------------------------------------
//		 TByteQueue::~TByteQueue
// ---------------------------------------------------------------------------

TByteQueue::~TByteQueue()
{
    delete fImplementation;
}


// ---------------------------------------------------------------------------
//		 TByteQueue::Put
// ---------------------------------------------------------------------------

void TByteQueue::Put( uae_u8 b )
{
    fImplementation->Put( b );
}


// ---------------------------------------------------------------------------
//		 TByteQueue::Get
// ---------------------------------------------------------------------------

uae_u8 TByteQueue::Get( void )
{
    return fImplementation->Get();
}


// ---------------------------------------------------------------------------
//		 TByteQueue::GetUsed
// ---------------------------------------------------------------------------

int TByteQueue::GetUsed( void )
{
    return fImplementation->GetUsed();
}


// ---------------------------------------------------------------------------
//		 TByteQueue::GetFree
// ---------------------------------------------------------------------------

int TByteQueue::GetFree( void )
{
    return fImplementation->GetFree();
}


// ---------------------------------------------------------------------------
//		 TByteQueue::WaitForDataAvailable
// ---------------------------------------------------------------------------

Bool TByteQueue::WaitForDataAvailable( int timeoutms )
{
    return fImplementation->WaitForDataAvailable( timeoutms );
}

// ===========================================================================
//		 TByteQueuePrivate
// ===========================================================================

// ---------------------------------------------------------------------------
//		 TByteQueuePrivate::TByteQueuePrivate
// ---------------------------------------------------------------------------

TByteQueuePrivate::TByteQueuePrivate():
        _available( &_mutex )
{
    _usedCount = 0;
    _freeCount = QUEUE_SIZE;
  
    _head = 0;
    _tail = 0;
}


// ---------------------------------------------------------------------------
//		 TByteQueuePrivate::~TByteQueuePrivate
// ---------------------------------------------------------------------------

TByteQueuePrivate::~TByteQueuePrivate()
{
}


// ---------------------------------------------------------------------------
//		 TByteQueue::Put
// ---------------------------------------------------------------------------

void TByteQueuePrivate::Put( uae_u8 b )
{
    omni_mutex_lock l(_mutex);

    // If the queue is full, just discard the data byte.  
    if (_usedCount >= QUEUE_SIZE)
    {
        return;
    }

    _buffer[_head] = b;
    _head = (_head + 1) % QUEUE_SIZE;
  
    _usedCount++;
    _freeCount--;
  
    // Tell clients that there may be new data in the buffer.
    _available.signal();
}


// ---------------------------------------------------------------------------
//		 TByteQueue::Get
// ---------------------------------------------------------------------------

uae_u8 TByteQueuePrivate::Get()
{
    omni_mutex_lock l(_mutex);

    // If the queue is empty, return a null byte (this shouldn't happen,
    // because the caller should always call GetUsed before Get).
  
    if (_usedCount == 0)
    {
        return 0;
    }
  
    uae_u8 r = _buffer[_tail];
    _tail = (_tail + 1) % QUEUE_SIZE;
  
    _freeCount++;
    _usedCount--;
  
    return r;
}


// ---------------------------------------------------------------------------
//		 TByteQueue::WaitForDataAvailable
// ---------------------------------------------------------------------------

Bool TByteQueuePrivate::WaitForDataAvailable( int timeoutms )
{
    // calc absolute time to wait until:
    unsigned long sec = 0, nanosec = 0;
    omni_thread::get_time( &sec, &nanosec, 0, timeoutms * 1000 );

  
    return _available.timedwait( sec, nanosec ) == 1;
}

// ---------------------------------------------------------------------------
//		 Others...
// ---------------------------------------------------------------------------

#if 0
static void InsertStr( char* dest, int insertBefore, const char* src )
{
}


static void FormatInteger( char* dest, uae_u32 integer )
{
}
#endif

