/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"

#include <FL/Fl.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Browser.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Check_Button.H>
#include <FL/Fl_Choice.H>
#include <FL/Fl_Input.H>
#include <FL/Fl_Light_Button.H>
#include <FL/Fl_Output.H>
#include <FL/Fl_Pixmap.H>
#include <FL/Fl_Return_Button.H>
#include <FL/Fl_Round_Button.H>
#include <FL/Fl_Tabs.H>
#include <FL/Fl_Widget.H>
#include <FL/Fl_Window.H>
#include <FL/fl_draw.H>
#include <FL/fl_file_chooser.H>
#include <string.h>

#include "Strings.r.h"
#include "Platform.h"
#include "ErrorHandling.h"
#include "AppPreferences.h"
#include "CPU_REG.h"

#include "MapFile.h"
#include "omnithread.h"
#include "fltk_Dialog.h"
#include "fltk_main.h"

#include "poser.xpm"

static omni_mutex sDialogMutex;
static omni_semaphore sDialogDone(0);
static char *sDialogTitle = NULL;
static char *sDialogMsg = NULL;
static int sDialogFlags;
static int sDialogResult;

// -----------------------------------------------------------------------------
// helper functions for getting data out of windows
#define SET_BOOL_FROM_BUTTON( bv, w, n ) bv = ((Fl_Button*)w->child(n))->value()==1?true:false

// -----------------------------------------------------------------------------
// callback used to close modal dialogs. modalResult is < 0 if not yet completed.
// set modalResult to be >=0 to complete dialog. Typically 0=cancel, 1=ok
static long modalResult;
static void modalCallback( Fl_Widget*, long int result )
{
    modalResult = result;
}

// -----------------------------------------------------------------------------
// modal dialog control loop
static int postModalDialog( Fl_Window *dlg )
{
    modalResult = -9999;
    dlg->show();

    // run our own event loop so that we have total control
    // of the UI thread
    for (;;)
    {
        Fl::wait();
        if ( modalResult != -9999 )
            break;
        if ( !dlg->visible() )
        {
            // user hit escape (cancel)
            modalResult = 0;
            break;
        }
    }
    dlg->hide();
    return modalResult;
}

// -----------------------------------------------------------------------------
// simple text output subclass which does word wrapping...
class WrapOutput : public Fl_Output
{
    bool _wrapped;
    string _unwrapped;
public:
    WrapOutput( int x, int y, int w, int h, char *l=NULL)
            : Fl_Output( x,y,w,h,l ), _wrapped(false)
        {}
    virtual void draw();
    virtual void resize( int x,int y,int w, int h ) { _wrapped = false; Fl_Output::resize(x,y,w,h); }

    void unwrappedValue( const string& s ) { _unwrapped = s; }
};

// -----------------------------------------------------------------------------
// wrap the text before drawing, give the current window size
void WrapOutput::draw()
{
    if ( !_wrapped )
    {
        int len = _unwrapped.length();
        char *inbuf = new char[ len + 1 ];
        char *outbuf = new char[ len + 1 ];

        strcpy( inbuf, _unwrapped.c_str() );

        int tw = w()+40;
        char *t = outbuf, *f = inbuf;
        char *curline = outbuf;
        char *pspace = NULL;

        // init out buffer to empty string
        *t = '\0';
        
        while ( *f )
        {
            // copy a char, and make sure the dest string is always null terminated
            *t = *f;
            *(t+1) = '\0';
            
            if ( *t == ' ' )
            {
                // if we see a space, remember it
                pspace = t;
            }
            else if ( *t == '\n' )
            {
                // if we see a newline, reset the space ptr
                // and resel the current line ptr.
                pspace = NULL;
                curline = t+1;
            }
            
            if ( (fl_width( curline ) > tw) && (pspace!=NULL) )
            {
                // if the width of the current line exceeds the window size,
                // and we have a previous space, then wrap.
                *pspace = '\n';
                curline = pspace + 1;
                pspace = NULL;
            }
            t++; f++;
        }

        value( outbuf );
        delete inbuf;
        delete outbuf;
        _wrapped = true;
    }
    Fl_Output::draw();
}

// -----------------------------------------------------------------------------
// create the message common dialog
Fl_Window* makeCommonDialog( const string& msg, const string& title,
                             const string buttons[], const int results[] )
{
    Fl_Window* w;
    {
        Fl_Window* o = new Fl_Window(343, 156, title.c_str());
        w = o;
        o->box(FL_EMBOSSED_BOX);
        {
            Fl_Return_Button* o = new Fl_Return_Button(250, 125, 85, 25, buttons[0].c_str());
            o->labelsize(12);
            o->callback( modalCallback, results[0] );
        }
        if ( !buttons[1].empty() )
        {
            Fl_Button* o = new Fl_Button(160, 125, 85, 25, buttons[1].c_str());
            o->labelsize(12);
            o->callback( modalCallback, results[1] );
        }
        if ( !buttons[2].empty() )
        {
            Fl_Button* o = new Fl_Button(70, 125, 85, 25, buttons[2].c_str());
            o->labelsize(12);
            o->callback( modalCallback, results[2] );
        }
        {
            WrapOutput* o = new WrapOutput(5, 5, 330,115);
            o->type(4);
            o->color(50);
            o->labelsize(10);
            o->textsize(10);
            o->align(132);
            Fl_Group::current()->resizable(o);
            o->unwrappedValue( msg );
        }
        o->set_modal();
        o->end();
    }
    return w;
}

// -----------------------------------------------------------------------------
// post the dialog (call in the context of the UI thread)
void fltkPostDialog( void )
{
    string buttons[3];
    int results[3];
    string label(" Note" );

    // get the mesage icon, and set it at appropriate:
    switch (sDialogFlags & Errors::kAlertMask)
    {
        case Errors::kNoteAlert:
            label = "Note";
            break;
        case Errors::kCautionAlert:
            label = "Caution";
            break;
        case Errors::kErrorAlert:
            label = "Error";
            break;
    }

    if (sDialogFlags & Errors::kOKCancelMask)
    {
        if ((sDialogFlags & Errors::kOKCancelMask) == Errors::kOK)
        {
            buttons[0] = Platform::GetString( kStr_OK );
            results[0] = Errors::kOK;
        }
        else if ((sDialogFlags & Errors::kOKCancelMask) == Errors::kCancel)
        {
            buttons[0] = Platform::GetString( kStr_Cancel );
            results[0] = Errors::kCancel;
        }
        else
        {
            buttons[0] = Platform::GetString( kStr_OK );
            results[0] = Errors::kOK;
            buttons[1] = Platform::GetString( kStr_Cancel );
            results[1] = Errors::kCancel;
        }
    }
    else if (sDialogFlags & Errors::kYesNoMask)
    {
        if ((sDialogFlags & Errors::kYesNoMask) == Errors::kYes)
        {
            buttons[0] = Platform::GetString( kStr_Yes );
            results[0] = Errors::kYes;
        }
        else if ((sDialogFlags & Errors::kYesNoMask) == Errors::kNo)
        {
            buttons[0] = Platform::GetString( kStr_No );
            results[0] = Errors::kNo;
        }
        else
        {
            buttons[0] = Platform::GetString( kStr_Yes );
            buttons[1] = Platform::GetString( kStr_No );
            results[0] = Errors::kYes;
            results[1] = Errors::kNo;
        }
    }
    else if (sDialogFlags & Errors::kContinueDebugResetMask)
    {
        int idx = 0;

        if (sDialogFlags & Errors::kContinue)
        {
            buttons[idx] = Platform::GetString( kStr_Continue );
            results[idx++] = Errors::kContinue;
        }

        if (sDialogFlags & Errors::kDebug)
        {
            buttons[idx] = Platform::GetString( kStr_Debug );
            results[idx++] = Errors::kDebug;
        }

        if (sDialogFlags & Errors::kReset)
        {
            buttons[idx] = Platform::GetString( kStr_Reset );
            results[idx++] = Errors::kReset;
        }
    }

    // post dialog:
    Fl_Window *dlg = makeCommonDialog( string(sDialogMsg), label, buttons, results );
    sDialogResult = postModalDialog( dlg );
    delete dlg;

    sDialogDone.post();
}

// -----------------------------------------------------------------------------
// tell the UI thread to post this dialog
int fltkDialog(const char* title, const char* msg, int flags)
{
    // remember the dialog parameters:
    sDialogFlags = flags;
    sDialogMsg = strdup( msg );
    sDialogTitle = strdup( title );
    
    if ( currentlyUIThread() )
    {
        fltkPostDialog();
    }
    else
    {
        // signal the UI thread to put up the given dialog:

        // acquire the dialog mutex. This will insure that only one thread is posting
        // a dialog request to the UI thread.
        sDialogMutex.lock();

        // Tell the UI thread that there is a dialog waiting to be posted:
        fltkExecuteUI( fltkPostDialog );

        // wait for the result
        sDialogDone.wait();

        // release the dialog mutex
        sDialogMutex.unlock();
    }

    // Free the string memory:
    if ( sDialogMsg )
    {
        free( sDialogMsg );
        sDialogMsg = NULL;
    }
    if ( sDialogTitle )
    {
        free( sDialogTitle );
        sDialogTitle = NULL;
    }

    return sDialogResult;  
}

// -----------------------------------------------------------------------------
// URL handler callback. Note that for this to work, netscape must currently
// be running.
//!TODO: work with other browsers, work even if netscape is note running
static void openURL( Fl_Widget* box, void *)
{
    char buffer[PATH_MAX];
    char url[PATH_MAX];

    // get the label, trim the opening '<' and closing '>'
    strcpy( url, &(box->label()[1]) );
    url[ strlen(url) -1 ] = '\0';
    
    sprintf( buffer, "netscape -remote 'openURL(%s)' || netscape '%s' &", url, url );
    system( buffer );
}


// -----------------------------------------------------------------------------
// construct the About Box window
Fl_Window* makeAboutWindow()
{
    Fl_Window* w;
    {
        Fl_Window* o = new Fl_Window(362, 231, "About PalmOS(tm) Emulator");
        w = o;
        o->box(FL_EMBOSSED_BOX);
        {
            Fl_Tabs* o = new Fl_Tabs(70, 10, 285, 180);
            {
                Fl_Group* o = new Fl_Group(95, 35, 260, 155, "Palm");
                o->labelsize(10);
                {
                    Fl_Box* o = new Fl_Box(110, 45, 240, 20, "PalmOS Emulator Version 2.1d28");
                    o->labelfont(1);
                    o->labelsize(12);
                    o->align(20);
                }
                {
                    Fl_Box* o = new Fl_Box(110, 70, 240, 15, "Copyright (c) 1998-1999 Palm Computing, Inc.,");
                    o->labelsize(10);
                    o->align(20);
                }
                {
                    Fl_Box* o = new Fl_Box(110, 85, 240, 15, "3Com Corporation or its subsidiaries");
                    o->labelsize(10);
                    o->align(20);
                }
                {
                    Fl_Button* o = new Fl_Button(110, 105, 115, 15, "<http://www.palm.com>");
                    o->box(FL_NO_BOX);
                    o->selection_color(1);
                    o->labelsize(10);
                    o->labelcolor(4);
                    o->callback((Fl_Callback*)openURL);
                    o->align(20);
                }
                {
                    Fl_Button* o = new Fl_Button(110, 120, 140, 15, "<mailto:devsupp@palm.com>");
                    o->box(FL_NO_BOX);
                    o->selection_color(1);
                    o->labelsize(10);
                    o->labelcolor(4);
                    o->callback((Fl_Callback*)openURL);
                    o->align(20);
                }
                {
                    Fl_Button* o = new Fl_Button(110, 155, 230, 15, "<http://www.palm.com/devzone/pose/seed.html>");
                    o->box(FL_NO_BOX);
                    o->selection_color(1);
                    o->labelsize(10);
                    o->labelcolor(4);
                    o->callback((Fl_Callback*)openURL);
                    o->align(20);
                }
                {
                    Fl_Box* o = new Fl_Box(110, 140, 165, 15, "Download the latest version from:");
                    o->labelsize(10);
                    o->align(20);
                }
                o->end();
            }
            {
                Fl_Group* o = new Fl_Group(95, 35, 260, 155, "Windows");
                o->labelsize(12);
                o->hide();
                {
                    Fl_Box* o = new Fl_Box(100, 55, 240, 15, "Original Windows Copilot by Greg Hewgill");
                    o->labelsize(10);
                    o->align(20);
                }
                {
                    Fl_Box* o = new Fl_Box(100, 70, 240, 15, "Portions copyright (c) 1996-1997");
                    o->labelsize(10);
                    o->align(20);
                }
                {
                    Fl_Button* o = new Fl_Button(100, 100, 125, 15, "<http://www.hewgill.com>");
                    o->box(FL_NO_BOX);
                    o->selection_color(1);
                    o->labelsize(10);
                    o->labelcolor(4);
                    o->callback((Fl_Callback*)openURL);
                    o->align(20);
                }
                {
                    Fl_Button* o = new Fl_Button(100, 115, 130, 15, "<mailto:greg@hewgill.com>");
                    o->box(FL_NO_BOX);
                    o->selection_color(1);
                    o->labelsize(10);
                    o->labelcolor(4);
                    o->callback((Fl_Callback*)openURL);
                    o->align(20);
                }
                o->end();
            }
            {
                Fl_Group* o = new Fl_Group(95, 35, 260, 155, "Macintosh");
                o->labelsize(12);
                o->hide();
                {
                    Fl_Box* o = new Fl_Box(100, 55, 240, 15, "Original Mac port by Craig Schofield");
                    o->labelsize(10);
                    o->align(20);
                }
                {
                    Fl_Box* o = new Fl_Box(100, 70, 240, 15, "Portions copyright (c) 1995-1997");
                    o->labelsize(10);
                    o->align(20);
                }
                {
                    Fl_Button* o = new Fl_Button(100, 100, 220, 15, "<http://members.aol.com/illumesoft/illume.html>");
                    o->box(FL_NO_BOX);
                    o->selection_color(1);
                    o->labelsize(10);
                    o->labelcolor(4);
                    o->callback((Fl_Callback*)openURL);
                    o->align(20);
                }
                {
                    Fl_Button* o = new Fl_Button(100, 115, 220, 15, "<mailto:ILLUMESoft@aol.com>");
                    o->box(FL_NO_BOX);
                    o->selection_color(1);
                    o->labelsize(10);
                    o->labelcolor(4);
                    o->callback((Fl_Callback*)openURL);
                    o->align(20);
                }
                o->end();
            }
            {
                Fl_Group* o = new Fl_Group(95, 35, 260, 155, "UAE");
                o->labelsize(12);
                o->hide();
                {
                    Fl_Box* o = new Fl_Box(100, 55, 240, 15, "UAE Amiga Emulator by Bernd Schmidt");
                    o->labelsize(10);
                    o->align(20);
                }
                {
                    Fl_Box* o = new Fl_Box(100, 70, 240, 15, "Portions copyright (c) 1995-1997");
                    o->labelsize(10);
                    o->align(20);
                }
                {
                    Fl_Button* o = new Fl_Button(100, 100, 220, 15, "<http://www.freiburg.linux.de/~uae/>");
                    o->box(FL_NO_BOX);
                    o->selection_color(1);
                    o->labelsize(10);
                    o->labelcolor(4);
                    o->callback((Fl_Callback*)openURL);
                    o->align(20);
                }
                {
                    Fl_Button* o = new Fl_Button(100, 115, 225, 15, "<mailto:crux@pool.informatik.rwth-aachen.de>");
                    o->box(FL_NO_BOX);
                    o->selection_color(1);
                    o->labelsize(10);
                    o->labelcolor(4);
                    o->callback((Fl_Callback*)openURL);
                    o->align(20);
                }
                o->end();
            }
            o->end();
        }
        {
            Fl_Box* o = new Fl_Box(10, 15, 50, 60, "");
            Fl_Pixmap* p = new Fl_Pixmap( poser_xpm );
            p->label( o );
        }
        {
            Fl_Return_Button* o = new Fl_Return_Button(280, 197, 75, 25, "OK");
            o->labelsize(12);
            o->callback( modalCallback, 1 );
            w->hotspot(o);
        }
        o->set_modal();
        o->end();
    }
    return w;
}

// -----------------------------------------------------------------------------
// post the about box window
void showAboutBox()
{
    Fl_Window *aboutWin = makeAboutWindow();
    postModalDialog( aboutWin );
    delete aboutWin;
}

// -----------------------------------------------------------------------------
// construct the Gremlins dialog
Fl_Window* makeGremlinsWindow()
{
	AppPreferences::GremlinInfo& info = gPrefs->GetGremlinInfo();

    Fl_Window* w;
    {
        Fl_Window* o = new Fl_Window(447, 223, "New Gremlin");
        w = o;
        o->box(FL_ENGRAVED_BOX);
        o->color(50);
        o->labelcolor(50);
        o->align(129);
        {
            char buffer[16];
            Fl_Input* o = new Fl_Input(165, 20, 60, 25, "Gremlin Number (0-999):");
            o->labelsize(10);
            o->textsize(10);
            o->type(FL_INT_INPUT);

            sprintf( buffer, "%ld", info.fNumber);
            o->value( buffer );
        }
        {
            char buffer[16];
            Fl_Input* o = new Fl_Input(165, 50, 60, 25, "Number of Steps (-1 = forever):");
            o->labelsize(10);
            o->textsize(10);
            o->type(FL_INT_INPUT);

            sprintf( buffer, "%ld", info.fSteps);
            o->value( buffer );
        }
        {
            char buffer[16];
            Fl_Input* o = new Fl_Input(165, 80, 60, 25, "Snapshot Frequency (0 = never):");
            o->labelsize(10);
            o->textsize(10);
            o->type(FL_INT_INPUT);

            sprintf( buffer, "%ld", info.fSaveFrequency);
            o->value( buffer );
        }
        {
            Fl_Light_Button* o = new Fl_Light_Button(10, 120, 215, 25, "Continue past warnings");
            o->labelsize(10);
            o->value( info.fContinuePastWarnings? 1 : 0 );
        }
        {
            Fl_Browser* o = new Fl_Browser(235, 20, 200, 155, "Choose application to launch:");
            o->type(3);
            o->labelsize(10);
            o->textsize(10);
            o->align(1);

            bool selected = false;
            AppInfoList appList;
            GetApplicationNames( appList );
			for ( unsigned int ii = 0; ii < appList.size (); ++ii )
			{
                o->add( appList[ii].name );

				// If the item we just added is in our "selected" list,
				// then select it.
				
				AppInfoList::iterator iter = info.fAppList.begin();
				while ( iter != info.fAppList.end() )
				{
					if ( appList[ii] == *iter++ )
					{
						o->select( ii+1 );
						selected = true;
					}
				}
                if ( !selected )
                    o->select( 1 );
            }
        }
        {
            Fl_Return_Button* o = new Fl_Return_Button(365, 185, 70, 25, "OK");
            o->labelsize(10);
            o->callback( modalCallback, 1 );
            w->hotspot(o);
        }
        {
            Fl_Button* o = new Fl_Button(290, 185, 70, 25, "Cancel");
            o->callback( modalCallback, 0 );
            o->labelsize(12);
        }
        o->set_modal();
        o->end();
    }
    return w;
}

// -----------------------------------------------------------------------------
// post the gremlins window

#define ID_GREMLIN_NUMBER				0
#define ID_GREMLIN_EVENTS				1
#define ID_GREMLIN_SAVE_FREQUENCY		2
#define ID_GREMLIN_WARNINGS				3
#define ID_GREMLIN_APPLICATIONS			4
	
bool showNewGremlinWindow()
{
    Fl_Window *gremlinWin = makeGremlinsWindow();
    bool result = false;

    if ( postModalDialog( gremlinWin ) == 1 )
    {
        // scan the values out of the input widgets:
        AppPreferences::GremlinInfo& info = gPrefs->GetGremlinInfo();
        
        Fl_Input *i = (Fl_Input*)gremlinWin->child(ID_GREMLIN_NUMBER);
        sscanf(i->value(), "%ld", &info.fNumber );
        
        i = (Fl_Input*)gremlinWin->child(ID_GREMLIN_EVENTS);
        sscanf(i->value(), "%ld", &info.fSteps );
        
        i = (Fl_Input*)gremlinWin->child(ID_GREMLIN_SAVE_FREQUENCY);
        sscanf(i->value(), "%ld", &info.fSaveFrequency );
        
        SET_BOOL_FROM_BUTTON( info.fContinuePastWarnings, gremlinWin, ID_GREMLIN_WARNINGS );

        Fl_Browser* br = (Fl_Browser*)gremlinWin->child(ID_GREMLIN_APPLICATIONS);
        info.fAppList.clear();

        AppInfoList appList;
        GetApplicationNames( appList );
        for (int ii = 1; ii <= br->size(); ++ii)
        {
            if ( br->selected( ii ) )
                info.fAppList.push_back( appList[ ii-1 ] );
        }
        
        result = true;
    }

    delete gremlinWin;
    return result;
}

// -----------------------------------------------------------------------------
// functions to convert between popup menu indecies and configuration enums
// and vice versa
static int ramSizeToIndex( int size )
{
    switch ( size )
    {
        case 128: return 0;
        case 256: return 1;
        case 512: return 2;
        case 1024: return 3;
        case 2048: return 4;
        case 3072: return 5;
        case 4096: return 6;
        case 8192: return 7;
    }
    return 0;
}

static int indexToRamSize( int idx )
{
    switch ( idx )
    {
        case 0: return 128;
        case 1: return 256;
        case 2: return 512;
        case 3: return 1024;
        case 4: return 2048;
        case 5: return 3072;
        case 6: return 4096;
        case 7: return 8192;
    }
    return 0;
}

static int silkToIndex( SilkscreenType type )
{
    switch ( type )
    {
        case kSilkscreenEnglish: return 0;
        case kSilkscreenJapanese: return 1;
		default: assert (false);
    }
    return 0;
}

static SilkscreenType indexToSilk( int idx )
{
    switch ( idx )
    {
        case 0: return kSilkscreenEnglish;
        case 1: return kSilkscreenJapanese;
    }
    return kSilkscreenEnglish;
}

static int deviceToIndex( DeviceType type )
{
    switch ( type )
    {
        case kDevicePilot1000: return 0;
        case kDevicePalmPilotPersonal: return 1;
        case kDevicePalmIII: return 2;
        case kDevicePalmIIIx: return 3;
        case kDevicePalmV: return 4;
        case kDevicePalmVII: return 5;
		default: assert (false);
    }
    return 2;
}

static DeviceType indexToDevice( int idx )
{
    switch ( idx )
    {
        case 0: return kDevicePilot1000;
        case 1: return kDevicePalmPilotPersonal;
        case 2: return kDevicePalmIII;
        case 3: return kDevicePalmIIIx;
        case 4: return kDevicePalmV;
        case 5: return kDevicePalmVII;
    }
    return kDevicePalmIII;
}

// -----------------------------------------------------------------------------
// open the choose rom image dialog
void chooseROMCallback( Fl_Widget* widget, void* param )
{
    FileReference *f = (FileReference*) param;
    Fl_Button *btn = (Fl_Button*)widget;

    // ask the user to find a rom file:
    char *fname = fl_file_chooser( "Choose a ROM image",
                                   "*.{rom,ROM}",
                                   f->GetFilePath() );

    if ( fname != NULL )
    {
        FileReference *nf = new FileReference( fname );
        btn->user_data( nf );
        btn->label( nf->GetFileName().c_str() );
        delete f;
    }
}

// -----------------------------------------------------------------------------
// new emulator dialog
Fl_Window* makeNewEmulatorWindow( const AppPreferences::Configuration& cfg )
{
    Fl_Window* w;
    {
        Fl_Window* o = new Fl_Window(246, 160, "New Configuration");
        w = o;
        o->box(FL_ENGRAVED_BOX);
        {
            Fl_Choice* o = new Fl_Choice(60, 10, 175, 20, "Device:");
            o->add( "Pilot (1000 or 5000)|PalmPilot (Personal or Pro)|Palm III|Palm IIIx|Palm V|Palm VII" );

            o->value( deviceToIndex( cfg.fDeviceType ) );
            o->labelsize(10);
            o->textsize(10);
        }
        {
            Fl_Choice* o = new Fl_Choice(60, 35, 115, 20, "Silkscreen:");
            o->add( "English|Japanese" );
            o->value( silkToIndex( cfg.fSilkscreenType ) );
            o->labelsize(10);
            o->textsize(10);
        }
        {
            Fl_Choice* o = new Fl_Choice(60, 60, 115, 20, "RAM Size:");
            o->add( "128KB|256KB|512KB|1024KB (1MB)|2048KB (2 MB)|3072KB (3 MB)|4096KB (4 MB)|8192KB (8 MB)");
            o->value( ramSizeToIndex( cfg.fRAMSize ) );
            o->labelsize(10);
            o->textsize(10);
        }
        {
            Fl_Button* o = new Fl_Button(60, 85, 175, 20, cfg.fROMFile.GetFileName().c_str() );
            o->callback( chooseROMCallback, new FileReference( cfg.fROMFile ) );
            o->labelsize(10);
        }
        {
            Fl_Box* o = new Fl_Box(10, 85, 50, 25, "ROM File:");
            o->labelsize(10);
        }
        {
            Fl_Button* o = new Fl_Button(100, 125, 65, 25, "Cancel");
            o->labelsize(12);
            o->callback( modalCallback, 0 );
        }
        {
            Fl_Return_Button* o = new Fl_Return_Button(170, 125, 65, 25, "Ok");
            o->labelsize(12);
            o->callback( modalCallback, 1 );
            w->hotspot(o);
        }
        o->set_modal();
        o->end();
    }
    return w;
}

// -----------------------------------------------------------------------------
// post the new emulator window
bool showNewEmulatorWindow( AppPreferences::Configuration& cfg )
{
    Fl_Window *newWin = makeNewEmulatorWindow( cfg );
    bool result = false;
    
    if ( postModalDialog( newWin ) == 1 )
    {
        Fl_Choice* c = (Fl_Choice*)newWin->child(0);
        cfg.fDeviceType = indexToDevice( c->value() );

        c = (Fl_Choice*)newWin->child(1);
        cfg.fSilkscreenType = indexToSilk( c->value() );

        c = (Fl_Choice*)newWin->child(2);
        cfg.fRAMSize = indexToRamSize( c->value() );

        FileReference *f = (FileReference*) newWin->child(3)->user_data();
        cfg.fROMFile = FileReference( *f );
        delete f;

        result = true;
    }

    delete newWin;
    return result;
}

// -----------------------------------------------------------------------------
// callback to choose a new skin
static void chooseSkinCallback( Fl_Widget*, void* data )
{
    char *fname = fl_file_chooser( "Choose a skin file", "*.skin", NULL );
    
    if ( fname != NULL )
    {
        gHostPrefs->fSkinFiles.push_back( string( fname ) );
        Fl_Choice *c = (Fl_Choice*) ((Fl_Window*)data)->child(6);

        MapFile map( fname );
        char *name = map.find( "Name" );
        if ( name != NULL )
            c->add( name );
        else
            c->add( fname );

        // choose the one we just found ( -2: one is for the menu label
        // which is hidden, and one is for zero-based to one-based conversion )
        c->value( c->size() -2 );
    }
}

// -----------------------------------------------------------------------------
// create the preferences dialog
Fl_Window* makePrefsWindow( Preferences& prefs )
{
    Fl_Window* w;
    {
        Fl_Window* o = new Fl_Window(244, 373, "Preferences");
        w = o;
        {
            Fl_Box* o = new Fl_Box(10, 25, 225, 65, "Communications");
            o->box(FL_THIN_DOWN_FRAME);
            o->labelfont(1);
            o->labelsize(12);
            o->align(5);
        }
        {
            Fl_Box* o = new Fl_Box(10, 115, 225, 55, "Display");
            o->box(FL_THIN_DOWN_FRAME);
            o->labelfont(1);
            o->labelsize(12);
            o->align(5);
        }
        {
            Fl_Box* o = new Fl_Box(10, 195, 225, 75, "Closing / Quitting");
            o->box(FL_THIN_DOWN_FRAME);
            o->labelfont(1);
            o->labelsize(12);
            o->align(5);
        }
        {
            Fl_Box* o = new Fl_Box(10, 295, 225, 30, "HotSync User Name");
            o->box(FL_THIN_DOWN_FRAME);
            o->labelfont(1);
            o->labelsize(12);
            o->align(5);
        }
        {
            Fl_Input* o = new Fl_Input(70, 35, 160, 20, "Serial Port:");
            o->value( prefs._serialDevice.c_str() );
            o->labelsize(10);
            o->textsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(15, 60, 190, 20, "Redirect NetLIb calls to host TCP/IP");
            o->down_box(FL_THIN_DOWN_BOX);
            o->labelsize(10);
            o->value( prefs.RedirectNetLib()? 1 : 0 );
        }
        {
            Fl_Choice* o = new Fl_Choice(45, 132, 115, 20, "Skin:");

            gHostPrefs->statSkinFileList();
            ParamList::iterator skinIter;
            int n=0;
            for ( skinIter=gHostPrefs->fSkinFiles.begin(); skinIter != gHostPrefs->fSkinFiles.end(); n++ )
            {
                string skinFile = *skinIter++;
                MapFile map( skinFile.c_str() );
                char *name = map.find( "Name" );
                if ( name != NULL )
                    o->add( name );
                else
                    o->add( skinFile.c_str() );
                if ( strcmp( skinFile.c_str(), gHostPrefs->getSkinFile() ) == 0 )
                {
                    o->value( n );
                }
            }
            o->labelsize(10);
            o->textsize(10);
        }
        {
            Fl_Button* o = new Fl_Button(175, 132, 45, 20, "Other...");
            o->labelsize(10);
            o->callback( chooseSkinCallback, w );
        }
        {
            Fl_Group* o = new Fl_Group(15, 195, 120, 70);
            {
                Fl_Round_Button* o = new Fl_Round_Button(15, 200, 120, 25, "Always save session");
                o->type(102);
                o->down_box(FL_ROUND_DOWN_BOX);
                o->labelsize(10);
                if ( prefs.SaveAlways() )
                    o->value(1);
            }
            {
                Fl_Round_Button* o = new Fl_Round_Button(15, 220, 115, 25, "Ask to save session");
                o->type(102);
                o->down_box(FL_ROUND_DOWN_BOX);
                o->labelsize(10);
                if ( prefs.SaveAsk() )
                    o->value(1);
            }
            {
                Fl_Round_Button* o = new Fl_Round_Button(15, 240, 120, 25, "Never save session");
                o->type(102);
                o->down_box(FL_ROUND_DOWN_BOX);
                o->labelsize(10);
                if ( prefs.SaveNever() )
                    o->value(1);
            }
            o->end();
        }
        {
            Fl_Input* o = new Fl_Input(15, 300, 215, 20);
            o->labelsize(10);
            o->textsize(10);
            o->value( prefs.fUserName.c_str() );
        }
        {
            Fl_Button* o = new Fl_Button(100, 340, 65, 25, "Cancel");
            o->labelsize(12);
            o->callback( modalCallback, 0 );
        }
        {
            Fl_Return_Button* o = new Fl_Return_Button(170, 340, 65, 25, "Ok");
            o->labelsize(12);
            w->hotspot(o);
            o->callback( modalCallback, 1 );
        }
        o->set_modal();
        o->end();
    }
    return w;
}

// -----------------------------------------------------------------------------
// post the prefs window. This is a member of the Preferences object so that
// we can have intimate access to get/set the members
bool Preferences::editPreferences()
{
    Fl_Window *prefsWin = makePrefsWindow( *this );
    bool result = false;
    
    if ( postModalDialog( prefsWin ) == 1 )
    {
        Fl_Input* i = (Fl_Input*)prefsWin->child(4);
        _serialDevice = i->value();

        SET_BOOL_FROM_BUTTON( fRedirectNetLib, prefsWin, 5 );

        Fl_Choice *c = (Fl_Choice*) prefsWin->child(6);
        ParamList::iterator skinIter;
        int n=0;
        for ( skinIter=gHostPrefs->fSkinFiles.begin(); skinIter != gHostPrefs->fSkinFiles.end(); n++ )
        {
            string skinFile = *skinIter++;
            if ( n == c->value() )
            {
                _skinFilePath = skinFile;
                break;
            }
        }
        
        Fl_Button *b = (Fl_Button*)((Fl_Group*)prefsWin->child(8))->child(0);
        if ( b->value() == 1 )
            fCloseAction = kSaveAlways;
        else
        {
            b = (Fl_Button*)((Fl_Group*)prefsWin->child(8))->child(1);
            fCloseAction = b->value() == 1? kSaveAsk : kSaveNever;
        }

        i = (Fl_Input*)prefsWin->child(9);
        fUserName = i->value();
        
        // changes
        result = true;
    }

    delete prefsWin;
    return result;
}

// -----------------------------------------------------------------------------
// create the debugging preferences dialog
Fl_Window* makeDebugWindow( Preferences& prefs )
{
    Fl_Window* w;
    {
        Fl_Window* o = new Fl_Window(327, 162, "Debugging Options");
        w = o;
        o->box(FL_EMBOSSED_BOX);
        {
            Fl_Check_Button* o = new Fl_Check_Button(15, 10, 120, 20, "Low-Memory Access");
            o->down_box(FL_THIN_DOWN_BOX);
            o->value( prefs.fReportLowMemoryAccess? 1 : 0 );
            o->labelsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(15, 25, 125, 20, "System Globals Access");
            o->down_box(FL_THIN_DOWN_BOX);
            o->value( prefs.fReportSystemGlobalAccess? 1 : 0 );
            o->labelsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(15, 40, 90, 20, "Screen Access");
            o->down_box(FL_THIN_DOWN_BOX);
            o->value( prefs.fReportScreenAccess? 1 : 0 );
            o->labelsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(15, 55, 140, 20, "Hardware Register Access");
            o->down_box(FL_THIN_DOWN_BOX);
            o->value( prefs.fReportHardwareRegisterAccess? 1 : 0 );
            o->labelsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(15, 70, 130, 20, "MemMgr Data Structure");
            o->down_box(FL_THIN_DOWN_BOX);
            o->value( prefs.fReportMemMgrDataAccess? 1 : 0 );
            o->labelsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(15, 85, 120, 20, "Storage Heap Access");
            o->down_box(FL_THIN_DOWN_BOX);
            o->value( prefs.fReportStorageHeapAccess? 1 : 0 );
            o->labelsize(10);
            o->deactivate();
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(15, 100, 120, 20, "MemMgr Semaphore");
            o->down_box(FL_THIN_DOWN_BOX);
            o->value( prefs.fReportMemMgrSemaphore? 1 : 0 );
            o->labelsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(175, 10, 105, 20, "Low Stack Access");
            o->down_box(FL_THIN_DOWN_BOX);
            o->value( prefs.fReportLowStackAccess? 1 : 0 );
            o->labelsize(10);
            o->deactivate();
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(175, 25, 110, 20, "Free Chunk Access");
            o->down_box(FL_THIN_DOWN_BOX);
            o->value( prefs.fReportFreeChunkAccess? 1 : 0 );
            o->labelsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(175, 40, 130, 20, "Unlocked Chunk Access");
            o->down_box(FL_THIN_DOWN_BOX);
            o->value( prefs.fReportUnlockedChunkAccess? 1 : 0 );
            o->labelsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(175, 55, 135, 20, "Unitialized Stack Access");
            o->down_box(FL_THIN_DOWN_BOX);
            o->value( prefs.fReportUninitializedStackAccess? 1 : 0 );
            o->labelsize(10);
            o->deactivate();
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(175, 70, 145, 20, "Uninitialized Chunk Access");
            o->down_box(FL_THIN_DOWN_BOX);
            o->value( prefs.fReportUninitializedChunkAccess? 1 : 0 );
            o->labelsize(10);
            o->deactivate();
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(175, 85, 130, 20, "Stack Almost Overflow");
            o->down_box(FL_THIN_DOWN_BOX);
            o->value( prefs.fReportStackAlmostOverflow? 1 : 0 );
            o->labelsize(10);
        }
        {
            Fl_Return_Button* o = new Fl_Return_Button(245, 130, 70, 25, "OK");
            o->labelsize(12);
            o->callback( modalCallback, 1 );
            w->hotspot(o);
        }
        {
            Fl_Button* o = new Fl_Button(170, 130, 70, 25, "Cancel");
            o->labelsize(12);
            o->callback( modalCallback, 0);
        }
        o->set_modal();
        o->end();
    }
    return w;
}

// -----------------------------------------------------------------------------
// post the debug prefs window. This is a member of the Preferences object so that
// we can have intimate access to get/set the members
bool Preferences::editDebugging()
{
    Fl_Window *debugWin = makeDebugWindow( *this );
    bool result = false;
    
    if ( postModalDialog( debugWin ) == 1 )
    {
        SET_BOOL_FROM_BUTTON( fReportLowMemoryAccess, debugWin, 0 );
        SET_BOOL_FROM_BUTTON( fReportSystemGlobalAccess, debugWin, 1 );
        SET_BOOL_FROM_BUTTON( fReportScreenAccess, debugWin, 2 );
        SET_BOOL_FROM_BUTTON( fReportHardwareRegisterAccess, debugWin, 3 );
        SET_BOOL_FROM_BUTTON( fReportMemMgrDataAccess, debugWin, 4 );
        SET_BOOL_FROM_BUTTON( fReportStorageHeapAccess, debugWin, 5 );
        SET_BOOL_FROM_BUTTON( fReportMemMgrSemaphore, debugWin, 6 );
        SET_BOOL_FROM_BUTTON( fReportLowStackAccess, debugWin, 7 );
        SET_BOOL_FROM_BUTTON( fReportFreeChunkAccess, debugWin, 8 );
        SET_BOOL_FROM_BUTTON( fReportUnlockedChunkAccess, debugWin, 9 );
        SET_BOOL_FROM_BUTTON( fReportUninitializedStackAccess, debugWin, 10 );
        SET_BOOL_FROM_BUTTON( fReportUninitializedChunkAccess, debugWin, 11 );
        SET_BOOL_FROM_BUTTON( fReportStackAlmostOverflow, debugWin, 12 );

        // changes
        result = true;
    }

    delete debugWin;
    return result;
}

// -----------------------------------------------------------------------------
// helper data functions for getting & setting prefs & checkboxes

//! TODO: get rid of these globals!
static int curPanel;
static Fl_Window *curLogWin;
static Preferences *curPrefs;

void updatePreference( Fl_Window* w, int btnId, uae_u8& value, int bitMask )
{
    Fl_Check_Button *btn = (Fl_Check_Button*) w->child( btnId );
	if ( btn->value() != 0 )
		value |= bitMask;
	else
		value &= ~bitMask;
}

void updateCheckButton( Fl_Window* w, int btnId, uae_u8 value, int bitMask )
{
    Fl_Check_Button *btn = (Fl_Check_Button*) w->child( btnId );
	btn->value( ((value & bitMask) != 0) ? 1 : 0 );
}

void updateCheckButtons( Fl_Window* w, Preferences& prefs, int panel )
{
	int bitMask = (panel == 1)? AppPreferences::kNormalLogging
        : AppPreferences::kGremlinLogging;

	updateCheckButton( w, 4, prefs.fLogErrorMessages, bitMask );
	updateCheckButton( w, 5, prefs.fLogWarningMessages, bitMask );
	updateCheckButton( w, 6, prefs.fLogGremlins, bitMask );
	updateCheckButton( w, 7, prefs.fLogCPUOpcodes, bitMask );
	updateCheckButton( w, 8, prefs.fLogEnqueuedEvents, bitMask );
	updateCheckButton( w, 9, prefs.fLogDequeuedEvents, bitMask );
	updateCheckButton( w, 10, prefs.fLogSystemCalls, bitMask );
	updateCheckButton( w, 11, prefs.fLogApplicationCalls, bitMask );
	updateCheckButton( w, 12, prefs.fLogSerial, bitMask );
	updateCheckButton( w, 13, prefs.fLogNetLib, bitMask );
	updateCheckButton( w, 14, prefs.fLogExgMgr, bitMask );
	updateCheckButton( w, 15, prefs.fLogHLDebugger, bitMask );
	updateCheckButton( w, 16, prefs.fLogLLDebugger, bitMask );
	updateCheckButton( w, 17, prefs.fLogSerialData, bitMask );
	updateCheckButton( w, 18, prefs.fLogNetLibData, bitMask );
	updateCheckButton( w, 19, prefs.fLogExgMgrData, bitMask );
	updateCheckButton( w, 20, prefs.fLogHLDebuggerData, bitMask );
	updateCheckButton( w, 21, prefs.fLogLLDebuggerData, bitMask );
}

void updatePreferences( Fl_Window* w, Preferences& prefs, int panel )
{
	int bitMask = (panel == 1)? AppPreferences::kNormalLogging
        : AppPreferences::kGremlinLogging;

	updatePreference( w, 4, prefs.fLogErrorMessages, bitMask );
	updatePreference( w, 5, prefs.fLogWarningMessages, bitMask );
	updatePreference( w, 6, prefs.fLogGremlins, bitMask );
	updatePreference( w, 7, prefs.fLogCPUOpcodes, bitMask );
	updatePreference( w, 8, prefs.fLogEnqueuedEvents, bitMask );
	updatePreference( w, 9, prefs.fLogDequeuedEvents, bitMask );
	updatePreference( w, 10, prefs.fLogSystemCalls, bitMask );
	updatePreference( w, 11, prefs.fLogApplicationCalls, bitMask );
	updatePreference( w, 12, prefs.fLogSerial, bitMask );
	updatePreference( w, 13, prefs.fLogNetLib, bitMask );
	updatePreference( w, 14, prefs.fLogExgMgr, bitMask );
	updatePreference( w, 15, prefs.fLogHLDebugger, bitMask );
	updatePreference( w, 16, prefs.fLogLLDebugger, bitMask );
	updatePreference( w, 17, prefs.fLogSerialData, bitMask );
	updatePreference( w, 18, prefs.fLogNetLibData, bitMask );
	updatePreference( w, 19, prefs.fLogExgMgrData, bitMask );
	updatePreference( w, 20, prefs.fLogHLDebuggerData, bitMask );
	updatePreference( w, 21, prefs.fLogLLDebuggerData, bitMask );
}

void panelCallback( Fl_Widget* w, long panel )
{
    if ( curPanel == panel )
        return;

    // save current panel:
    updatePreferences( curLogWin, *curPrefs, curPanel );

    // switch to new panel
    curPanel = (int) panel;
    updateCheckButtons( curLogWin, *curPrefs, curPanel );
}

// -----------------------------------------------------------------------------
// construct the logging preferences window:
Fl_Window* makeLoggingWindow()
{
    Fl_Window* w;
    {
        Fl_Window* o = new Fl_Window(356, 242, "Logging Options");
        w = o;
        {
            Fl_Group* o = new Fl_Group(220, 5, 130, 25);
            {
                Fl_Round_Button* o = new Fl_Round_Button(225, 5, 55, 25, "Normal");
                o->type(102);
                o->down_box(FL_ROUND_DOWN_BOX);
                o->value(1);
                o->labelsize(10);
                o->callback( panelCallback, 1 );
            }
            {
                Fl_Round_Button* o = new Fl_Round_Button(285, 5, 65, 25, "Gremlins");
                o->type(102);
                o->down_box(FL_ROUND_DOWN_BOX);
                o->labelsize(10);
                o->callback( panelCallback, 2 );
            }
            o->end();
        }
        {
            Fl_Box* o = new Fl_Box(15, 30, 335, 170, "Log:");
            o->box(FL_THIN_DOWN_FRAME);
            o->labelfont(1);
            o->labelsize(12);
            o->align(5);
        }
        {
            Fl_Return_Button* o = new Fl_Return_Button(280, 210, 70, 25, "OK");
            o->labelsize(12);
            o->callback( modalCallback, 1 );
            w->hotspot(o);
        }
        {
            Fl_Button* o = new Fl_Button(205, 210, 70, 25, "Cancel");
            o->labelsize(12);
            o->callback( modalCallback, 0 );
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(20, 35, 90, 20, "Error Messages");
            o->down_box(FL_THIN_DOWN_BOX);
            o->labelsize(10);
            o->deactivate();
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(20, 50, 105, 20, "Warning Messages");
            o->down_box(FL_THIN_DOWN_BOX);
            o->labelsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(20, 65, 100, 20, "Misc Gremlin Info");
            o->down_box(FL_THIN_DOWN_BOX);
            o->labelsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(20, 80, 110, 20, "Assembley Opcodes");
            o->down_box(FL_THIN_DOWN_BOX);
            o->labelsize(10);
            o->deactivate();
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(195, 35, 90, 20, "Posted Events");
            o->down_box(FL_THIN_DOWN_BOX);
            o->labelsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(195, 50, 105, 20, "Received Events");
            o->down_box(FL_THIN_DOWN_BOX);
            o->labelsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(195, 65, 75, 20, "System Calls");
            o->down_box(FL_THIN_DOWN_BOX);
            o->labelsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(195, 80, 95, 20, "Application Calls");
            o->down_box(FL_THIN_DOWN_BOX);
            o->labelsize(10);
            o->deactivate();
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(20, 110, 85, 20, "Serial Activity");
            o->down_box(FL_THIN_DOWN_BOX);
            o->labelsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(20, 125, 90, 20, "NetLib Activity");
            o->down_box(FL_THIN_DOWN_BOX);
            o->labelsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(20, 140, 95, 20, "ExgMgr Activity");
            o->down_box(FL_THIN_DOWN_BOX);
            o->labelsize(10);
            o->deactivate();
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(20, 155, 160, 20, "High-Level Debugger Activity");
            o->down_box(FL_THIN_DOWN_BOX);
            o->labelsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(20, 170, 160, 20, "Low-Level Debugger Activity");
            o->down_box(FL_THIN_DOWN_BOX);
            o->labelsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(195, 110, 70, 20, "Serial Data");
            o->down_box(FL_THIN_DOWN_BOX);
            o->labelsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(195, 125, 75, 20, "NetLib Data");
            o->down_box(FL_THIN_DOWN_BOX);
            o->labelsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(195, 140, 95, 20, "ExgMgr Data");
            o->down_box(FL_THIN_DOWN_BOX);
            o->labelsize(10);
            o->deactivate();
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(195, 155, 145, 20, "High-Level Debugger Data");
            o->down_box(FL_THIN_DOWN_BOX);
            o->labelsize(10);
        }
        {
            Fl_Check_Button* o = new Fl_Check_Button(195, 170, 145, 20, "Low-Level Debugger Data");
            o->down_box(FL_THIN_DOWN_BOX);
            o->labelsize(10);
        }
        o->set_modal();
        o->end();
    }
    return w;
}

// -----------------------------------------------------------------------------
// post the logging prefs window. This is a member of the Preferences object so that
// we can have intimate access to get/set the members
bool Preferences::editLogging()
{
    Fl_Window *loggingWin = makeLoggingWindow();

    // set up globals (yuck)
    curLogWin = loggingWin;
    curPrefs = new Preferences( *this );
    curPanel = 1;

    // get the info into the window the first time:
    updateCheckButtons( curLogWin, *curPrefs, 1 );
    
    bool result = false;
    if ( postModalDialog( loggingWin ) == 1 )
    {
        // save current panel into master prefs object
        updatePreferences( curLogWin, *this, curPanel );

        // switch to other panel
        curPanel = (curPanel == 1)? 2 : 1;
        updateCheckButtons( curLogWin, *curPrefs, curPanel );

        // save this panel into master prefs
        updatePreferences( curLogWin, *this, curPanel );
        
        // changes
        result = true;
    }

    curLogWin = NULL;
    delete curPrefs;

    delete loggingWin;
    return result;
}
